#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "../inc/buffer_management.h"
#include "../inc/device.h"
#include "../inc/generic_types.h"
#include "../inc/net_types.h"
#include "../inc/network.h"
#include "../inc/socket.h"
#include "../inc/utils.h"
#include "../inc/services/time_synchronisation.h"
#include  "../library/driver/ethernet/patmos_tsn_mac.h"

#include <machine/exceptions.h>
#include <machine/patmos.h>
#include <machine/rtc.h>
#include <machine/uart.h>

#define SOCKET_CNT 8
#define MESSAGES_PER_SIZE 100
#define MESSAGES_PER_ITERATION 4

// Network Stack config
#ifndef NETWORK_STACK_MEMORY_POOL_SIZE
    #define NETWORK_STACK_MEMORY_POOL_SIZE 0xE000
#endif
uint8_t network_stack_buffer[NETWORK_STACK_MEMORY_POOL_SIZE];

// Network interface config
char eth0_if_name[] = "eth0";
uint8_t eth0_if_ipv4[4] = {172, 16, 0, 201};
uint8_t eth0_if_mac[6] = {0x00, 0x80, 0x6e, 0xF0, 0xDA, 0x42};
uint16_t eth0_interface_id= 0xFFFF;

tsn_gcl_config_t tsn_config;
tsn_gcl_entry_t gcl[10];

// socket configuration
uint16_t socket_id = 0xFFFF;
unsigned short src_port = 54321;

// destination config
socket_address_t eth0_dest_addr;
//unsigned char dst_mac[6] ={0x00, 0x0e, 0xc6, 0xc9, 0x48, 0x59};
unsigned char eth0_dst_ipv4[4] = {172, 16, 0, 100};
unsigned short dst_port = 12345;
char data_str_short[] = "This is a test msg";
char data_str_long[] = "This is a very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very very vvery very very very very very very very very very very very very very very very massive test msg";

// Timer interrupt
unsigned long long next_tick_timestamp;
volatile unsigned long long int isr_runtime;
unsigned long long int isr_runtime_longest;
unsigned long long int isr_start_delay;
unsigned long long int isr_start_delay_longest;

// Time Synchronisation
time_synchronisation_config_t ts_conf;

volatile uint32_t packets_sent = 0;
volatile uint32_t packets_received = 0;

#define TICK_INTERVALL_US 10000
#define LEDS (*((volatile _IODEV unsigned *)  PATMOS_IO_LED))

int init_network_and_socket(void)
{
    buffer_management_config_t buffer_mgmt_config;
    buffer_mgmt_config.type = BUFFER_MGMT_TYPE_TLSF;
    buffer_mgmt_config.start_address = (void *) (network_stack_buffer);
    buffer_mgmt_config.end_address =  (void *) (network_stack_buffer + NETWORK_STACK_MEMORY_POOL_SIZE);
    buffer_mgmt_config.tlsf_cfg.fl_lowest_cls = 6;
    buffer_mgmt_config.tlsf_cfg.fl_cls_cnt = 10;
    buffer_mgmt_config.tlsf_cfg.sl_cls_cnt =2;

    // init network stack
    network_init(&buffer_mgmt_config);


    uint8_t var =0;
    if(network_set_option(0xFFFF, NETWORK_OPTION_LOGGING, &var, sizeof(var)) != NET_ERR_OK)
    {
        utils_log("Error in netowrk interface when setting option NETWORK_OPTION_LOGGING\n");
        return -1;
    }

    var =0;
    if(socket_set_option(GLOBAL_SOCKET, SOCKET_OPTION_DEBUG_LOG, &var, sizeof(var) ) != NET_ERR_OK)
    {
        utils_log("Error in socket when setting option SOCKET_OPTION_DEBUG_LOG\n");
        return -1;
    }

    var =0;
    if(socket_set_option(GLOBAL_SOCKET, SOCKET_OPTION_MISSING_L2_ADDRESS_NO_FAIL, &var, sizeof(var) ) != NET_ERR_OK)
    {
        utils_log("Error in socket when setting option SOCKET_OPTION_DEBUG_LOG\n");
        return -1;
    }

    return 0;
}

void init_eth0(void)
{
    network_interface_config_t if_config;
    
    if_config.interface_name = (char*) &eth0_if_name;
    if_config.interface_type = NETWORK_INTERFACE_PATMOS_ETHERNET_TSN;

    if_config.device_config.access_type = DEVICE_ACCESS_MEMORY_MAPPED_IO;
    if_config.device_config.interupt_ena = 1;
    if_config.device_config.interupt_id = 22; // interrupts start with exception ID 16, EthMac Intterupt = 6 => exception ID 22
    if_config.device_config.queue_length = 30;
    if_config.device_config.rx_buffers = 3;

    if_config.device_config.timer_cfg.interupt_ena=1;
    if_config.device_config.timer_cfg.interupt_id = 24; // interrupts start with exception ID 16, Timer use Intterupt = 11 => exception ID 27
    if_config.device_config.timer_cfg.timer_base_address = (void*) 0xf00f0000;

    if_config.device_config.buffer_mgmt_config.type = BUFFER_MGMT_TYPE_MMIO_TLSF;
    if_config.device_config.buffer_mgmt_config.start_address = (void *) (PATMOS_IO_ETH);
    if_config.device_config.buffer_mgmt_config.end_address =  (void *) (PATMOS_IO_ETH + 0xE000);
    if_config.device_config.buffer_mgmt_config.tlsf_cfg.fl_lowest_cls = 5;//4;
    if_config.device_config.buffer_mgmt_config.tlsf_cfg.fl_cls_cnt = 11;//12;
    if_config.device_config.buffer_mgmt_config.tlsf_cfg.sl_cls_cnt =2 ;

    if_config.device_config.mac_address.length = 6;
    if_config.device_config.mac_address.address[0] = eth0_if_mac[0];
    if_config.device_config.mac_address.address[1] = eth0_if_mac[1];
    if_config.device_config.mac_address.address[2] = eth0_if_mac[2];
    if_config.device_config.mac_address.address[3] = eth0_if_mac[3];
    if_config.device_config.mac_address.address[4] = eth0_if_mac[4];
    if_config.device_config.mac_address.address[5] = eth0_if_mac[5];
    
    if_config.device_config.mmio_config.start_register_address = (void *) (PATMOS_IO_ETH + 0xF000);
    // layer 2 config
    if_config.layer2_config.l2_type = NETWORK_L2_ETHERNET_VLAN;
    if_config.layer2_config.vlan.dei=0;
    if_config.layer2_config.vlan.pcp=0;
    if_config.layer2_config.vlan.vid=100;

    // network address configuration
    if_config.address.type = NETWORK_ADDRESS_IP;
    if_config.address.ip.ip_address.version = IP_VERSION_4;
    if_config.address.ip.ip_address.ipv4_addr[0] = eth0_if_ipv4[0];
    if_config.address.ip.ip_address.ipv4_addr[1] = eth0_if_ipv4[1];
    if_config.address.ip.ip_address.ipv4_addr[2] = eth0_if_ipv4[2];
    if_config.address.ip.ip_address.ipv4_addr[3] = eth0_if_ipv4[3];
    if_config.address.ip.netmask = 24;

    eth0_interface_id = network_register_interface(&if_config);
    if(eth0_interface_id == 0xFFFF)
    {
        utils_log("Error in interface registration\n");
    }
}

void init_socket_eth0(void)
{
   
    socket_id = socket(SOCKET_DOMAIN_IP, SOCKET_TYPE_DATAGRAM, 0);
    if(socket_id == 0xFFFF)
    {
        utils_log("Error in socket creation\n");
        return;
    } 
}

void init_dest_addr(void)
{
    eth0_dest_addr.ip_address.version = IP_VERSION_4;
    eth0_dest_addr.ip_address.ipv4_addr[0] = eth0_dst_ipv4[0];
    eth0_dest_addr.ip_address.ipv4_addr[1] = eth0_dst_ipv4[1];
    eth0_dest_addr.ip_address.ipv4_addr[2] = eth0_dst_ipv4[2];
    eth0_dest_addr.ip_address.ipv4_addr[3] = eth0_dst_ipv4[3];
    eth0_dest_addr.port = dst_port;
}

// a basic handler for faults
void fault_handler(void) {
    unsigned source = exc_get_source();
    //LEDS = source;

    switch(source) {
    case 0: puts("Illegal operation"); break;
    case 1: puts("Illegal memory access"); break;
    default:puts("FAULT"); break;
    }
    while (1)
    {
        /* code */
    }
}

void isr_routine(void)
{
    SEI();

    network_interface_get_message(eth0_interface_id);

    CLI();
}

void set_timer(void)
{
    uint8_t cnt = 0;
    do{
        next_tick_timestamp +=TICK_INTERVALL_US;
        cnt++;
    }
    while (next_tick_timestamp <= get_cpu_usecs());
    
    if(cnt > 1)
    {

#ifndef WCET_ANALYSIS
        puts("INTERRUPT WAS TOO LONG BLOCKED!!!");
#endif

        do{
            next_tick_timestamp +=TICK_INTERVALL_US;
            cnt++;
        }
        while (next_tick_timestamp <= get_cpu_usecs());
    }
    // rearm RTC Clock interrupt
    arm_usec_timer(next_tick_timestamp);
}

/**
 * @brief Handle timer tick
 *  
 * Interrupt callback function to handle timer ticks
 */

void handle_tick(void) __attribute__((naked));
void handle_tick()
{
    exc_prologue();
    
    isr_routine();

    set_timer();

    exc_epilogue();
}

int main(void)
{

    uint8_t*data_short = (uint8_t*) &data_str_short;
    uint16_t data_len_short =  strlen( (char*) &data_str_short);

    uint8_t*data_long = (uint8_t*) &data_str_long;
    uint16_t data_len_long =  strlen( (char*) &data_str_long);

    socket_address_t msg_src_addr;
    uint16_t data_length;
    network_error_t ret;
    

    uint32_t cnt=1;
    LEDS =0;

    for (unsigned i = 0; i < 32; i++) {
        exc_register(i, &fault_handler);
    }
    
    if(init_network_and_socket() != 0)
    {
        return -1;
    }

    init_eth0();    
    if(eth0_interface_id == 0xFFFF)
    {
        return -1;
    }
    
    init_socket_eth0();
    
    init_dest_addr();
    
#ifdef WCET_ANALYSIS
    handle_tick();
#endif


    next_tick_timestamp =get_cpu_usecs() + TICK_INTERVALL_US;
    //arm_usec_timer(next_tick_timestamp);

    // configure timer interrupt
    exc_register(17, &handle_tick); 

    intr_unmask_all();
    intr_clear_all_pending();
    // enable interrupts
    intr_enable();

    uint32_t packets_sent_last = 0;
    
    packets_received =0;
    packets_sent =0;
    
    time_interval_t offset, mean_path_delay;
    time_interval_t last_offset, last_mean_path_delay;

    unsigned long long int start_time = get_cpu_usecs();

    // Wait untel arp was successful
    ret = NET_ERR_ARP_UNKNOWN_ADDR;
    while (ret == NET_ERR_ARP_UNKNOWN_ADDR)
    {
        ret = socket_send(socket_id, &eth0_dest_addr, data_short, data_len_short );
        network_interface_get_message(eth0_interface_id);
    }
    
    printf("cycles send 64 byte;cycle send time 1500 byte;\n");
    for (uint32_t message_cnt=0; message_cnt < MESSAGES_PER_SIZE; message_cnt++)
    {
        uint64_t start_time_short = get_cpu_cycles();
        ret = socket_send(socket_id, &eth0_dest_addr, data_short, data_len_short );
        uint64_t end_time_short = get_cpu_cycles();
        uint64_t runtime_send_short = end_time_short-start_time_short;
        if(ret != NET_ERR_OK && ret != NET_ERR_ARP_UNKNOWN_ADDR && ret != NET_ERR_DRVR_FRAME_QUEUED && ret != NET_ERR_DRVR_TIMESLOT_TO_SHORT && ret != NET_ERR_DRVR_TSN_QUEUES_CLOSED)
        {
            CLI();
            utils_log("eth0 socket_send: 0x%04hx\n", ret);
            utils_log("ERROR HALTING\n", ret);
            while (1){}
        }
        
        uint64_t start_time_long = get_cpu_cycles();
        ret = socket_send(socket_id, &eth0_dest_addr, data_long, data_len_long );
        uint64_t end_time_long = get_cpu_cycles();
        uint64_t runtime_send_long = end_time_long-start_time_long;
        
        if(ret != NET_ERR_OK && ret != NET_ERR_ARP_UNKNOWN_ADDR && ret != NET_ERR_DRVR_FRAME_QUEUED && ret != NET_ERR_DRVR_TIMESLOT_TO_SHORT && ret != NET_ERR_DRVR_TSN_QUEUES_CLOSED)
        {
            CLI();
            utils_log("eth0 socket_send: 0x%04hx\n", ret);
            utils_log("ERROR HALTING\n", ret);
            while (1){}
        }
        printf("%llu;%llu;\n", runtime_send_short, runtime_send_long);
        
    }
        
    return 0;
}
