#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "../inc/buffer_management.h"
#include "../inc/device.h"
#include "../inc/generic_types.h"
#include "../inc/net_types.h"
#include "../inc/network.h"
#include "../inc/socket.h"
#include "../inc/utils.h"
#include "../inc/services/time_synchronisation.h"
#include  "../library/driver/ethernet/patmos_tsn_mac.h"

#include <machine/exceptions.h>
#include <machine/patmos.h>
#include <machine/rtc.h>
#include <machine/uart.h>

// Network Stack config
#ifndef NETWORK_STACK_MEMORY_POOL_SIZE
    #define NETWORK_STACK_MEMORY_POOL_SIZE 0xE000
#endif
uint8_t network_stack_buffer[NETWORK_STACK_MEMORY_POOL_SIZE];

// Network interface config
char eth0_if_name[] = "eth0";
uint8_t eth0_if_ipv4[4] = {172, 16, 0, 201};
uint8_t eth0_if_mac[6] = {0x00, 0x80, 0x6e, 0xF0, 0xDA, 0x42};
uint16_t eth0_interface_id= 0xFFFF;

char eth1_if_name[] = "eth1";
uint8_t eth1_if_ipv4[4] = {172, 16, 1, 201};
uint8_t eth1_if_mac[6] = {0x00, 0x80, 0x6e, 0xF0, 0xDA, 0x43};
uint16_t eth1_interface_id= 0xFFFF;

tsn_gcl_config_t tsn_config;
tsn_gcl_entry_t gcl[10];

// socket configuration
uint16_t eth0_socket_id= 0xFFFF;
uint16_t eth0_isr_socket_id= 0xFFFF;
unsigned short src_port = 54321;
unsigned short src_port_isr = 55555;

// Timer interrupt
unsigned long long next_tick_timestamp;
volatile unsigned long long int isr_runtime;
unsigned long long int isr_runtime_longest;
unsigned long long int isr_start_delay;
unsigned long long int isr_start_delay_longest;

// Time Synchronisation
time_synchronisation_config_t ts_conf;

#define CYCLE_INTERVAL_US 100000
#define TICK_INTERVALL_US 10000
#define LEDS (*((volatile _IODEV unsigned *)  PATMOS_IO_LED))

int init_network_and_socket(void)
{
    buffer_management_config_t buffer_mgmt_config;
    buffer_mgmt_config.type = BUFFER_MGMT_TYPE_TLSF;
    buffer_mgmt_config.start_address = (void *) (network_stack_buffer);
    buffer_mgmt_config.end_address =  (void *) (network_stack_buffer + NETWORK_STACK_MEMORY_POOL_SIZE);
    buffer_mgmt_config.tlsf_cfg.fl_lowest_cls = 6;
    buffer_mgmt_config.tlsf_cfg.fl_cls_cnt = 10;
    buffer_mgmt_config.tlsf_cfg.sl_cls_cnt =2;

    // init network stack
    network_init(&buffer_mgmt_config);


    uint8_t var =0;
    if(network_set_option(0xFFFF, NETWORK_OPTION_LOGGING, &var, sizeof(var)) != NET_ERR_OK)
    {
        utils_log("Error in netowrk interface when setting option NETWORK_OPTION_LOGGING\n");
        return -1;
    }

    var =0;
    if(socket_set_option(GLOBAL_SOCKET, SOCKET_OPTION_DEBUG_LOG, &var, sizeof(var) ) != NET_ERR_OK)
    {
        utils_log("Error in socket when setting option SOCKET_OPTION_DEBUG_LOG\n");
        return -1;
    }

    var =0;
    if(socket_set_option(GLOBAL_SOCKET, SOCKET_OPTION_MISSING_L2_ADDRESS_NO_FAIL, &var, sizeof(var) ) != NET_ERR_OK)
    {
        utils_log("Error in socket when setting option SOCKET_OPTION_DEBUG_LOG\n");
        return -1;
    }

    return 0;
}

void init_eth0(void)
{
    network_interface_config_t if_config;
    
    if_config.interface_name = (char*) &eth0_if_name;
    if_config.interface_type = NETWORK_INTERFACE_PATMOS_ETHERNET_TSN;

    if_config.device_config.access_type = DEVICE_ACCESS_MEMORY_MAPPED_IO;
    if_config.device_config.interupt_ena = 1;
    if_config.device_config.interupt_id = 22; // interrupts start with exception ID 16, EthMac Intterupt = 6 => exception ID 22
    if_config.device_config.queue_length = 30;
    if_config.device_config.rx_buffers = 3;

    if_config.device_config.timer_cfg.interupt_ena=1;
    if_config.device_config.timer_cfg.interupt_id = 24; // interrupts start with exception ID 16, Timer use Intterupt = 11 => exception ID 27
    if_config.device_config.timer_cfg.timer_base_address = (void*) 0xf00f0000;

    if_config.device_config.buffer_mgmt_config.type = BUFFER_MGMT_TYPE_MMIO_TLSF;
    if_config.device_config.buffer_mgmt_config.start_address = (void *) (PATMOS_IO_ETH);
    if_config.device_config.buffer_mgmt_config.end_address =  (void *) (PATMOS_IO_ETH + 0xE000);
    if_config.device_config.buffer_mgmt_config.tlsf_cfg.fl_lowest_cls = 5;//4;
    if_config.device_config.buffer_mgmt_config.tlsf_cfg.fl_cls_cnt = 11;//12;
    if_config.device_config.buffer_mgmt_config.tlsf_cfg.sl_cls_cnt =2 ;

    if_config.device_config.mac_address.length = 6;
    if_config.device_config.mac_address.address[0] = eth0_if_mac[0];
    if_config.device_config.mac_address.address[1] = eth0_if_mac[1];
    if_config.device_config.mac_address.address[2] = eth0_if_mac[2];
    if_config.device_config.mac_address.address[3] = eth0_if_mac[3];
    if_config.device_config.mac_address.address[4] = eth0_if_mac[4];
    if_config.device_config.mac_address.address[5] = eth0_if_mac[5];
    
    if_config.device_config.mmio_config.start_register_address = (void *) (PATMOS_IO_ETH + 0xF000);
    // layer 2 config
    if_config.layer2_config.l2_type = NETWORK_L2_ETHERNET;

    // network address configuration
    if_config.address.type = NETWORK_ADDRESS_IP;
    if_config.address.ip.ip_address.version = IP_VERSION_4;
    if_config.address.ip.ip_address.ipv4_addr[0] = eth0_if_ipv4[0];
    if_config.address.ip.ip_address.ipv4_addr[1] = eth0_if_ipv4[1];
    if_config.address.ip.ip_address.ipv4_addr[2] = eth0_if_ipv4[2];
    if_config.address.ip.ip_address.ipv4_addr[3] = eth0_if_ipv4[3];
    if_config.address.ip.netmask = 24;

    eth0_interface_id = network_register_interface(&if_config);
    if(eth0_interface_id == 0xFFFF)
    {
        utils_log("Error in interface registration\n");
    }
}

void init_eth1(void)
{
    network_interface_config_t if_config;
    
    if_config.interface_name = (char*) &eth1_if_name;
    if_config.interface_type = NETWORK_INTERFACE_PATMOS_ETHERNET_TSN;

    if_config.device_config.access_type = DEVICE_ACCESS_MEMORY_MAPPED_IO;
    if_config.device_config.interupt_ena = 0;
    if_config.device_config.queue_length = 100;
    if_config.device_config.rx_buffers = 5;
    if_config.device_config.timer_cfg.interupt_ena=0;

    if_config.device_config.buffer_mgmt_config.type = BUFFER_MGMT_TYPE_MMIO_TLSF;
    if_config.device_config.buffer_mgmt_config.start_address = (void *) (PATMOS_IO_ETH1);
    if_config.device_config.buffer_mgmt_config.end_address =  (void *) (PATMOS_IO_ETH1 + 0xE000);
    if_config.device_config.buffer_mgmt_config.tlsf_cfg.fl_lowest_cls = 5;//4;
    if_config.device_config.buffer_mgmt_config.tlsf_cfg.fl_cls_cnt = 11;//12;
    if_config.device_config.buffer_mgmt_config.tlsf_cfg.sl_cls_cnt =2 ;

    if_config.device_config.mac_address.length = 6;
    if_config.device_config.mac_address.address[0] = eth1_if_mac[0];
    if_config.device_config.mac_address.address[1] = eth1_if_mac[1];
    if_config.device_config.mac_address.address[2] = eth1_if_mac[2];
    if_config.device_config.mac_address.address[3] = eth1_if_mac[3];
    if_config.device_config.mac_address.address[4] = eth1_if_mac[4];
    if_config.device_config.mac_address.address[5] = eth1_if_mac[5];
    
    if_config.device_config.mmio_config.start_register_address = (void *) (PATMOS_IO_ETH1 + 0xF000);
    // layer 2 config
    if_config.layer2_config.l2_type = NETWORK_L2_ETHERNET;

    // network address configuration
    if_config.address.type = NETWORK_ADDRESS_IP;
    if_config.address.ip.ip_address.version = IP_VERSION_4;
    if_config.address.ip.ip_address.ipv4_addr[0] = eth1_if_ipv4[0];
    if_config.address.ip.ip_address.ipv4_addr[1] = eth1_if_ipv4[1];
    if_config.address.ip.ip_address.ipv4_addr[2] = eth1_if_ipv4[2];
    if_config.address.ip.ip_address.ipv4_addr[3] = eth1_if_ipv4[3];
    if_config.address.ip.netmask = 24;

    eth1_interface_id = network_register_interface(&if_config);
    if(eth1_interface_id == 0xFFFF)
    {
        utils_log("Error in interface registration\n");
    }
}

static void setup_time_sync_config(void)
{
    ts_conf.netw_if = eth0_interface_id;
    ts_conf.domainNumber=0;
    
    ts_conf.announceTimeout = 4;
    ts_conf.logAnnounceInterval = 2;
   
    ts_conf.logDelayReqInterval = 0;
    ts_conf.logDelayRespTimeout = 2;

    ts_conf.logSyncInterval = 1;

    ts_conf.clock_accuracy = PTP_CLOCK_ACCURACY_2_5_MS;
    ts_conf.forced_state = TIME_SYNC_FORCED_STATE_SLAVE;
}

// a basic handler for faults
void fault_handler(void) {
    unsigned source = exc_get_source();
    //LEDS = source;

    switch(source) {
    case 0: puts("Illegal operation"); break;
    case 1: puts("Illegal memory access"); break;
    default:puts("FAULT"); break;
    }
    while (1)
    {
        /* code */
    }
}

void isr_routine(void)
{
    SEI();
    static uint8_t cnt=0;
    cnt++;
    
    if(cnt& 0x1 ==1)
    {
        network_interface_get_message(eth0_interface_id);
    }
    /*
    else 
    {
        time_synchronisation_update();
    }
    
    */
    CLI();
}

void set_timer(void)
{
    uint8_t cnt = 0;
    do{
        next_tick_timestamp +=TICK_INTERVALL_US;
        cnt++;
    }
    while (next_tick_timestamp <= get_cpu_usecs());
    
    if(cnt > 1)
    {

#ifndef WCET_ANALYSIS
        puts("INTERRUPT WAS TOO LONG BLOCKED!!!");
#endif

        do{
            next_tick_timestamp +=TICK_INTERVALL_US;
            cnt++;
        }
        while (next_tick_timestamp <= get_cpu_usecs());
    }
    // rearm RTC Clock interrupt
    arm_usec_timer(next_tick_timestamp);
}

/**
 * @brief Handle timer tick
 *  
 * Interrupt callback function to handle timer ticks
 */

void handle_tick(void) __attribute__((naked));
void handle_tick()
{
    exc_prologue();
    
    isr_routine();

    set_timer();

    exc_epilogue();
}

int main(void)
{    
    
    network_error_t ret;
    LEDS =0;

    for (unsigned i = 0; i < 32; i++) {
        exc_register(i, &fault_handler);
    }
    
    if(init_network_and_socket() != 0)
    {
        return -1;
    }

    init_eth0();    
    if(eth0_interface_id == 0xFFFF)
    {
        return -1;
    }   

    init_eth1();    
    if(eth1_interface_id == 0xFFFF)
    {
        return -1;
    }   
    
    setup_time_sync_config();
    ret = time_synchronisation_init(& ts_conf);
    if(ret != NET_ERR_OK)
    {
        utils_log("time_synchronisation_init: 0x%04hx\n", ret);
        LEDS = 0x55;
        while(1){}
        return -1;
    }
    
    next_tick_timestamp =get_cpu_usecs() + TICK_INTERVALL_US;
    arm_usec_timer(next_tick_timestamp);

    // configure timer interrupt
    exc_register(17, &handle_tick); 

    intr_unmask_all();
    intr_clear_all_pending();
    // enable interrupts
    intr_enable();

    unsigned long long int start_time = get_cpu_usecs();
    unsigned long long int cycle_time = get_cpu_usecs();

    uint32_t ts_size;
    time_interval_t offset, last_offset, mean_path_delay;
    time_stamp_t eth0_ts, eth1_ts;
    time_interval_t drift;

    start_time = get_cpu_usecs();
    
    offset.seconds =0;
    offset.nanoseconds=0;

    printf("\n");
    
    /*
    while (1)
    {
        for(uint16_t i=0; i<1000;i++)
        {
            ret = network_get_option(eth0_interface_id , NETWORK_OPTION_GET_PTP_CLOCK, &(clock_measurement[i]), &ts_size);

            start_time = get_cpu_usecs();
            while (get_cpu_usecs()- start_time > 50){}

        }
        for(uint16_t i=0; i<1000;i++)
        {
            printf("%llu.%09u\n", clock_measurement[i].seconds, clock_measurement[i].nanoseconds);
        }
    }*/

    while ((offset.seconds == 0 && offset.nanoseconds == 0) || offset.seconds != 0)
    {

        time_synchronisation_update();
        time_sync_get_runtime_data(&offset, &mean_path_delay);
        
        if(start_time + CYCLE_INTERVAL_US < get_cpu_usecs())
        {
            printf(":");
            fflush(stdout);
            start_time += CYCLE_INTERVAL_US;
        }
        
        //while (start_time - get_cpu_usecs() < start_time){}
    }
    printf("\n");

    ret = network_get_option(eth0_interface_id , NETWORK_OPTION_GET_PTP_CLOCK, &eth0_ts, &ts_size);
    // add WCET compensation to the timestamp
    // network_set_option takes at worst 1872 cycles(23400 ns)
    eth0_ts.nanoseconds += 23400;

    if (eth0_ts.nanoseconds >= 1000000000)
    {
        eth0_ts.seconds++;
        eth0_ts.nanoseconds -= 1000000000;
    }

    ret = network_set_option(eth1_interface_id , NETWORK_OPTION_SET_PTP_CLOCK, &eth0_ts, &ts_size);
    
    printf("offset;mean_path_delay;eth0_ts;eth1_ts;drift;\n");

    start_time = get_cpu_usecs();
    cycle_time = start_time;
    while (1)
    {       
        time_synchronisation_update();
        time_sync_get_runtime_data(&offset, &mean_path_delay);
        //printf("%llu;", (cycle_time-start_time)/ 1000000);



        
        if(offset.nanoseconds != 0 && (offset.seconds != last_offset.seconds || offset.nanoseconds != last_offset.nanoseconds))
        {
            ret = network_get_option(eth0_interface_id , NETWORK_OPTION_GET_PTP_CLOCK, &eth0_ts, &ts_size);
            ret = network_get_option(eth1_interface_id , NETWORK_OPTION_GET_PTP_CLOCK, &eth1_ts, &ts_size);
            utils_time_stamp_sub(&eth0_ts, &eth1_ts, &drift);

            /*
            drift.nanoseconds = eth0_ts.nanoseconds;
            drift.seconds = eth0_ts.seconds;

            drift.nanoseconds -= eth1_ts.nanoseconds;
            drift.seconds -= eth1_ts.seconds;

            if ((drift.seconds != 0 && drift.nanoseconds < 0) ||
                (drift.nanoseconds <= -1000000000))
            {
                drift.seconds--;
                drift.nanoseconds += 1000000000;
            }
            */ 
            // network_get_option takes at worst 2271 cycles(28387.5 ns)
            drift.nanoseconds -= 28388;
            utils_time_interval_correct_of_uf(&drift);

            /*
            if ((drift.seconds != 0 && drift.nanoseconds < 0) ||
                (drift.nanoseconds <= -1000000000))
            {
                drift.seconds--;
                drift.nanoseconds += 1000000000;
            }
            */

            time_sync_get_runtime_data(&offset, &mean_path_delay);
            
            if(offset.nanoseconds <0 )
            {
                printf("-%lld.%09d;", offset.seconds, -offset.nanoseconds);
            }
            else
            {
                printf(" %lld.%09d;", offset.seconds, offset.nanoseconds);
            }
            
            printf("%lld.%09d;", mean_path_delay.seconds, mean_path_delay.nanoseconds);
            printf("%llu.%09u;", eth0_ts.seconds, eth0_ts.nanoseconds);
            printf("%llu.%09u;", eth1_ts.seconds, eth1_ts.nanoseconds);

            if(drift.nanoseconds <0 )
            {
                printf("-%lld.%09d;", drift.seconds, -drift.nanoseconds);
            }
            else
            {
                printf(" %lld.%09d;", drift.seconds, drift.nanoseconds);
            }

            printf("\n");
            cycle_time += CYCLE_INTERVAL_US;    
            while (cycle_time - get_cpu_usecs() < cycle_time){}
            
            last_offset.nanoseconds = offset.nanoseconds;
            last_offset.seconds = offset.seconds;
        }
    }

    return 0;
}
