#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <machine/patmos.h>
#include <machine/rtc.h>

#include "../inc/generic_types.h"
#include "../inc/buffer_management.h"
#include "../library/buffer_management/mmio/tlsf.h"

#define MAX_BUFFER_SIZE 56872
#define REPETITION_COUNT 100
#define FIRST_FLI 5
#define LAST_FLI 16
#define SL_CLS_COUNT 2
int main(int argc, char **argv)
{

    buffer_management_config_t buffer_mgmt_config;
    buffer_mgmt_config.type = BUFFER_MGMT_TYPE_MMIO_TLSF;
    buffer_mgmt_config.start_address = (void *) (PATMOS_IO_ETH + 0x0000);
    buffer_mgmt_config.end_address = (void *) (PATMOS_IO_ETH + 0xE000);
    buffer_mgmt_config.tlsf_cfg.fl_lowest_cls = FIRST_FLI; // 5
    buffer_mgmt_config.tlsf_cfg.fl_cls_cnt =  LAST_FLI - FIRST_FLI; // 11
    buffer_mgmt_config.tlsf_cfg.sl_cls_cnt =SL_CLS_COUNT; // 4

    uint64_t malloc_measurements[(LAST_FLI-FIRST_FLI)*(1 <<SL_CLS_COUNT)];
    uint64_t free_measurements[(LAST_FLI-FIRST_FLI)*(1 <<SL_CLS_COUNT)];

    uint16_t buffer_mgmt_id = buffer_management_register(&buffer_mgmt_config);

    printf("\nNo split/merge\n");

    for(uint8_t i=0; i<2;i++)
    {
        if(i==0)
        {
            printf("malloc;");
        }
        else
        {
            printf("free;");
        }
        for(uint16_t j=0; j < LAST_FLI-FIRST_FLI; j++)
        {
            uint16_t last_sls= 1 <<SL_CLS_COUNT; 
            if(j + FIRST_FLI  == LAST_FLI-1)
            {
                // last two classes are out of memory bounds
                last_sls-=2;
            }

            for(uint16_t k=0; k < last_sls; k++)
            {
                printf("fli %u sli %u;", j,k);
                fflush(stdout);
            }    
        }
    }

    printf("\n");

    for(uint16_t i=0; i < REPETITION_COUNT; i++)
    {
        uint16_t cnt =0;
        for(uint16_t j=FIRST_FLI; j < LAST_FLI; j++)
        {
            uint16_t fl_class_size_bottom =	(1 << j);
            uint16_t last_sls= 1 <<SL_CLS_COUNT; 
            if(j == LAST_FLI-1)
            {
                // last two classes are out of memory bounds
                last_sls-=2;
            }

            for(uint16_t k=0; k < last_sls; k++)
            {
                uint16_t start_sl_buffer_size = fl_class_size_bottom + ( k << (SL_CLS_COUNT+ j+1-FIRST_FLI)) ; 
                uint16_t end_sl_buffer_size = fl_class_size_bottom + ((k+1) << (SL_CLS_COUNT+ j+1-FIRST_FLI))-1; 
                uint16_t sl_size = end_sl_buffer_size - start_sl_buffer_size;
                sl_size = rand() % sl_size;
                uint16_t buffer_size =end_sl_buffer_size; // use last available buffer size s.t. no splitting will occur
                if(buffer_size > MAX_BUFFER_SIZE)
                {
                    buffer_size = MAX_BUFFER_SIZE;
                }

                void* buffer1 = buffer_management_malloc(sizeof(void *), buffer_mgmt_id); // allocate minimal buffer to surround desired buffer 
                void* buffer2 = buffer_management_malloc(end_sl_buffer_size + 1, buffer_mgmt_id); // allocate buffer of next larger class s.t. this buffer will be allocated without splitting
                void* buffer3 = buffer_management_malloc(sizeof(void *), buffer_mgmt_id); // allocate minimal buffer to surround desired buffer 
                if(buffer1 == NULL || buffer2 == NULL || buffer3 ==NULL )
                {
                    printf("1 ALARM!!!!!!!!\n");
                    while (1)
                    {
                    }
                    
                }
                buffer_management_free(buffer2, buffer_mgmt_id);

                uint64_t malloc_usecs_start =  get_cpu_cycles();
                void* buffer = buffer_management_malloc(buffer_size, buffer_mgmt_id);
                uint64_t malloc_usecs_end =  get_cpu_cycles();

                if(buffer == NULL)
                {
                    printf("2 ALARM!!!!!!!!\n");
                    while (1)
                    {
                    }
                    
                }

                uint64_t free_usecs_start =  get_cpu_cycles();
                buffer_management_free(buffer, buffer_mgmt_id);
                uint64_t free_usecs_end =  get_cpu_cycles();

                buffer_management_free(buffer1, buffer_mgmt_id);
                buffer_management_free(buffer3, buffer_mgmt_id);

                malloc_measurements[cnt] =  malloc_usecs_end- malloc_usecs_start;
                free_measurements[cnt] = free_usecs_end- free_usecs_start;
                cnt++;
            }    
        }

        for(uint8_t i=0; i<2;i++)
        {
            printf(";");
            uint16_t cnt =0;
            for(uint16_t j=0; j < LAST_FLI-FIRST_FLI; j++)
            {
                uint16_t last_sls= 1 <<SL_CLS_COUNT; 
                if(j + FIRST_FLI  == LAST_FLI-1)
                {
                    // last two classes are out of memory bounds
                    last_sls-=2;
                }

                for(uint16_t k=0; k < last_sls; k++)
                {
                    if(i==0)
                    {
                        printf("%llu;",malloc_measurements[cnt]);
                    }
                    else
                    {
                        printf("%llu;",free_measurements[cnt]);
                    }
                    fflush(stdout);
            
                    cnt++;
                }    
            }
        }

        printf("\n");
    }

    printf("\nOne sided split/merge\n");

    for(uint8_t i=0; i<2;i++)
    {
        if(i==0)
        {
            printf("malloc;");
        }
        else
        {
            printf("free;");
        }
        for(uint16_t j=0; j < LAST_FLI-FIRST_FLI; j++)
        {
            uint16_t last_sls= 1 <<SL_CLS_COUNT; 
            if(j + FIRST_FLI == LAST_FLI-1)
            {
                // last two classes are out of memory bounds
                last_sls-=2;
            }

            for(uint16_t k=0; k < last_sls; k++)
            {
                printf("fli %u sli %u;", j,k);
                fflush(stdout);
            }    
        }
    }
    printf("\n");

    for(uint16_t i=0; i < REPETITION_COUNT; i++)
    {
        uint16_t cnt =0;
        for(uint16_t j=FIRST_FLI; j < LAST_FLI; j++)
        {
            uint16_t fl_class_size_bottom =	(1 << j);

            uint16_t last_sls= 1 <<SL_CLS_COUNT; 
            if(j == LAST_FLI-1)
            {
                // last two classes are out of memory bounds
                last_sls-=2;
            }

            for(uint16_t k=0; k < last_sls; k++)
            {
                uint16_t start_sl_buffer_size = fl_class_size_bottom + ( k << (SL_CLS_COUNT+ j+1-FIRST_FLI)) ; 
                uint16_t end_sl_buffer_size = fl_class_size_bottom + ((k+1) << (SL_CLS_COUNT+ j+1-FIRST_FLI))-1; 
                uint16_t sl_size = end_sl_buffer_size - start_sl_buffer_size;
                sl_size = rand() % sl_size;
                uint16_t buffer_size = start_sl_buffer_size + sl_size;
                if(buffer_size > MAX_BUFFER_SIZE)
                {
                    buffer_size = MAX_BUFFER_SIZE;
                }

                uint64_t malloc_usecs_start =  get_cpu_cycles();
                void* buffer = buffer_management_malloc(buffer_size, buffer_mgmt_id);
                uint64_t malloc_usecs_end =  get_cpu_cycles();
                if(buffer == NULL)
                {
                    printf("ALARM!!!!!!!!\n");
                    while (1)
                    {
                        /* code */
                    }
                    
                }

                uint64_t free_usecs_start =  get_cpu_cycles();
                buffer_management_free(buffer, buffer_mgmt_id);
                uint64_t free_usecs_end =  get_cpu_cycles();

                malloc_measurements[cnt] =  malloc_usecs_end- malloc_usecs_start;
                free_measurements[cnt] = free_usecs_end- free_usecs_start;
                cnt++;
            }    
        }

        for(uint8_t i=0; i<2;i++)
        {
            printf(";");
            
            uint16_t cnt =0;
            for(uint16_t j=0; j < LAST_FLI-FIRST_FLI; j++)
            {
                uint16_t last_sls= 1 <<SL_CLS_COUNT; 
                if(j + FIRST_FLI  == LAST_FLI-1)
                {
                    // last two classes are out of memory bounds
                    last_sls-=2;
                }

                for(uint16_t k=0; k < last_sls; k++)
                {
                    if(i==0)
                    {
                        printf("%llu;",malloc_measurements[cnt]);
                    }
                    else
                    {
                        printf("%llu;",free_measurements[cnt]);
                    }
                    fflush(stdout);
            
                    cnt++;
                }    
            }
        }

        printf("\n");
    }

    printf("\nDouble sided split/merge\n");    
    for(uint8_t i=0; i<2;i++)
    {
        if(i==0)
        {
            printf("malloc;");
        }
        else
        {
            printf("free;");
        }
        for(uint16_t j=0; j < LAST_FLI-FIRST_FLI; j++)
        {
            uint16_t last_sls= 1 <<SL_CLS_COUNT; 
            if(j + FIRST_FLI  == LAST_FLI-1)
            {
                // last two classes are out of memory bounds
                last_sls-=2;
            }

            for(uint16_t k=0; k < last_sls; k++)
            {
                printf("fli %u sli %u;", j,k);
                fflush(stdout);
            }    
        }
    }
    
    printf("\n");

    for(uint16_t i=0; i < REPETITION_COUNT; i++)
    {
        uint16_t cnt =0;
        for(uint16_t j=FIRST_FLI; j < LAST_FLI; j++)
        {
            uint16_t fl_class_size_bottom =	(1 << j);

            uint16_t last_sls= 1 <<SL_CLS_COUNT; 
            if(j == LAST_FLI-1)
            {
                // last two classes are out of memory bounds
                last_sls-=2;
            }

            for(uint16_t k=0; k < last_sls; k++)
            {
                uint16_t start_sl_buffer_size = fl_class_size_bottom + ( k << (SL_CLS_COUNT+ j+1-FIRST_FLI)) ; 
                uint16_t end_sl_buffer_size = fl_class_size_bottom + ((k+1) << (SL_CLS_COUNT+ j+1-FIRST_FLI))-1; 
                uint16_t sl_size = end_sl_buffer_size - start_sl_buffer_size;
                sl_size = rand() % sl_size;
                uint16_t buffer_size = start_sl_buffer_size + sl_size;
                if(buffer_size > MAX_BUFFER_SIZE)
                {
                    buffer_size = MAX_BUFFER_SIZE;
                }
                
                void* buffer1 = buffer_management_malloc(2*sizeof(void *), buffer_mgmt_id); // allocate minimal buffer ahead of actual buffer
                if(buffer1 == NULL)
                {
                    printf("ALARM!!!!!!!!\n");
                    while (1)
                    {
                        /* code */
                    }
                    
                }

                uint64_t malloc_usecs_start =  get_cpu_cycles();
                void* buffer = buffer_management_malloc(buffer_size, buffer_mgmt_id);
                uint64_t malloc_usecs_end =  get_cpu_cycles();

                if(buffer == NULL)
                {
                    printf("ALARM!!!!!!!!\n");
                    while (1)
                    {
                        /* code */
                    }
                    
                }

                void* buffer2 = buffer_management_malloc(2*sizeof(void *), buffer_mgmt_id); // allocate minimal buffer ahead of actual buffer
                void* buffer3 = buffer_management_malloc(2*sizeof(void *), buffer_mgmt_id); // allocate minimal buffer ahead of actual buffer
                

                // free buffers 1 and 2 before s.t. double sided merge is forced with varouis sizes
                buffer_management_free(buffer1, buffer_mgmt_id);
                buffer_management_free(buffer2, buffer_mgmt_id);

                uint64_t free_usecs_start =  get_cpu_cycles();
                buffer_management_free(buffer, buffer_mgmt_id);
                uint64_t free_usecs_end =  get_cpu_cycles();
                
                buffer_management_free(buffer3, buffer_mgmt_id);

                malloc_measurements[cnt] =  malloc_usecs_end- malloc_usecs_start;
                free_measurements[cnt] = free_usecs_end- free_usecs_start;
                cnt++;
            }    
        }

        for(uint8_t i=0; i<2;i++)
        {
            printf(";");
            
            uint16_t cnt =0;
            for(uint16_t j=0; j < LAST_FLI-FIRST_FLI; j++)
            {
                uint16_t last_sls= 1 <<SL_CLS_COUNT; 
                if(j + FIRST_FLI == LAST_FLI-1)
                {
                    // last two classes are out of memory bounds
                    last_sls-=2;
                }

                for(uint16_t k=0; k < last_sls; k++)
                {
                    if(i==0)
                    {
                        printf("%llu;",malloc_measurements[cnt]);
                    }
                    else
                    {
                        printf("%llu;",free_measurements[cnt]);
                    }
                    fflush(stdout);
            
                    cnt++;
                }    
            }
        }

        printf("\n");
    }

    return 0;
}
