#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <machine/patmos.h>
#include <machine/rtc.h>

#include "../inc/generic_types.h"
#include "../inc/buffer_management.h"
#include "../library/buffer_management/mmio/tlsf.h"

#define MAX_BUFFER_SIZE 56872
#define MIN_ALLOCATED_MEMORY 10000
#define MAX_BUFFER_COUNT 1000
#define FIRST_FLI 5
#define LAST_FLI 16
#define SL_CLS_COUNT 2
#define REPETITION_CNT 50000

int main(int argc, char **argv)
{

    buffer_management_config_t buffer_mgmt_config;
    buffer_mgmt_config.type = BUFFER_MGMT_TYPE_MMIO_TLSF;
    buffer_mgmt_config.start_address = (void *) (PATMOS_IO_ETH + 0x0000);
    buffer_mgmt_config.end_address = (void *) (PATMOS_IO_ETH + 0xE000);
    buffer_mgmt_config.tlsf_cfg.fl_lowest_cls = FIRST_FLI; // 5
    buffer_mgmt_config.tlsf_cfg.fl_cls_cnt =  LAST_FLI - FIRST_FLI; // 11
    buffer_mgmt_config.tlsf_cfg.sl_cls_cnt =SL_CLS_COUNT; // 4

    void *buffer_ptr_array[MAX_BUFFER_COUNT];
    uint16_t buffer_size_array[MAX_BUFFER_COUNT];

    for(uint16_t i=0; i< MAX_BUFFER_COUNT; i++)
    {
        buffer_ptr_array[i] = NULL;
        buffer_size_array[i] = 0;
        
    }

    uint16_t buffer_mgmt_id = buffer_management_register(&buffer_mgmt_config);

    printf("Total Memory: %lu\n", MAX_BUFFER_SIZE + sizeof(mmio_tlsf_block_header_t));
    printf("n;bytes_allocated;buffer_size;malloc;free;\n");

    uint64_t cnt=0;
    uint32_t allocated_memory = 0;

    while (cnt < REPETITION_CNT)
    {
        uint32_t operation = rand() %3;
        if(operation == 0 || allocated_memory < MIN_ALLOCATED_MEMORY )
        {
            uint16_t buffer_size = (rand() %(MAX_BUFFER_SIZE - 1)) + 1;
            uint64_t malloc_cycles_start =  get_cpu_cycles();
            void* buffer = buffer_management_malloc(buffer_size, buffer_mgmt_id);
            uint64_t malloc_cycles_end =  get_cpu_cycles();

            if(buffer == NULL)
            {
                // not enough free space for this buffer
                continue;
            }
            ptr_t buffer_address = (ptr_t) buffer;
            buffer_address -= sizeof(mmio_tlsf_block_header_t); 
            mmio_tlsf_block_header_free_ptr_t buffer_bh = (mmio_tlsf_block_header_free_ptr_t )buffer_address;
            
            uint16_t  buffer_id = rand() % MAX_BUFFER_COUNT;
            while(buffer_size_array[buffer_id] != 0)
            {
                buffer_id = rand() % MAX_BUFFER_COUNT;
            }

            buffer_size_array[buffer_id] = buffer_bh->bh.size << 2;
            buffer_ptr_array[buffer_id] = buffer;

            allocated_memory += buffer_size_array[buffer_id] + sizeof(mmio_tlsf_block_header_t);

            uint64_t malloc_cycles_duration = malloc_cycles_end- malloc_cycles_start;
            uint64_t malloc_cycles_duration_ns = malloc_cycles_duration*12.5; // cycle duration at 80 MHz

            printf("%llu;", cnt);
            printf("%u;", allocated_memory);
            printf("%u;", buffer_size_array[buffer_id]);
            printf("%llu;", malloc_cycles_duration_ns);
            printf(";\n");
            cnt++;
            
        }
        else if(operation == 1 && allocated_memory > 0)
        {
            uint16_t buffer_id = rand() % MAX_BUFFER_COUNT;
            while(buffer_size_array[buffer_id] == 0)
            {
                buffer_id = rand() % MAX_BUFFER_COUNT;
            }
            
            allocated_memory -= (buffer_size_array[buffer_id] + sizeof(mmio_tlsf_block_header_t));

            uint64_t free_cycles_start =  get_cpu_cycles();
            buffer_management_free( buffer_ptr_array[buffer_id], buffer_mgmt_id);
            uint64_t free_cycles_end =  get_cpu_cycles();
            uint64_t free_cycles_duration = free_cycles_end- free_cycles_start;
            uint64_t free_cycles_duration_ns = free_cycles_duration*12.5; // cycle duration at 80 MHz
            
            printf("%llu;", cnt);
            printf("%u;", allocated_memory);
            printf("%u;", buffer_size_array[buffer_id]);
            printf(";"); 
            printf("%llu;\n", free_cycles_duration_ns);
            cnt++;

            buffer_size_array[buffer_id] = 0;
        }   
        else
        {
            // do nothing
        }     
    }
    printf("end of prog\n");
    return 0;
}
