#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "../inc/buffer_management.h"
#include "../inc/device.h"
#include "../inc/generic_types.h"
#include "../inc/net_types.h"
#include "../inc/network.h"
#include "../inc/socket.h"
#include "../inc/utils.h"
#include  "../library/driver/ethernet/patmos_eth_mac.h"

#include <machine/exceptions.h>
#include <machine/patmos.h>
#include <machine/rtc.h>
#include <machine/uart.h>

// Network Stack config
#ifndef NETWORK_STACK_MEMORY_POOL_SIZE
    #define NETWORK_STACK_MEMORY_POOL_SIZE 0xE000
#endif
uint8_t network_stack_buffer[NETWORK_STACK_MEMORY_POOL_SIZE];

// Network interface config
char eth0_if_name[] = "eth0";
uint8_t eth0_if_ipv4[4] = {172, 16, 0, 201};
uint8_t eth0_if_mac[6] = {0x00, 0x80, 0x6e, 0xF0, 0xDA, 0x42};
uint16_t eth0_interface_id= 0xFFFF;

// socket configuration
uint16_t eth0_socket_id= 0xFFFF;
uint16_t eth0_isr_socket_id= 0xFFFF;
unsigned short src_port = 54321;
unsigned short src_port_isr = 55555;


// destination config
socket_address_t eth0_dest_addr;
//unsigned char dst_mac[6] ={0x00, 0x0e, 0xc6, 0xc9, 0x48, 0x59};
unsigned char eth0_dst_ipv4[4] = {172, 16, 0, 200};
unsigned short dst_port = 12345;
char hello_short[] = "hello ISR";
char hello_long[] = "hello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!"; //\thello world!";

// Timer interrupt
unsigned long long next_tick_timestamp;
volatile unsigned long long int isr_runtime;
unsigned long long int isr_runtime_longest;
unsigned long long int isr_start_delay;
unsigned long long int isr_start_delay_longest;

volatile uint32_t packets_sent = 0;
volatile uint32_t packets_received = 0;

volatile uint8_t isr_fired=0;
#define TICK_INTERVALL_US 10000
#define LEDS (*((volatile _IODEV unsigned *)  PATMOS_IO_LED))

int init_network_and_socket(void)
{
    buffer_management_config_t buffer_mgmt_config;
    buffer_mgmt_config.type = BUFFER_MGMT_TYPE_TLSF;
    buffer_mgmt_config.start_address = (void *) (network_stack_buffer);
    buffer_mgmt_config.end_address =  (void *) (network_stack_buffer) + NETWORK_STACK_MEMORY_POOL_SIZE;
    buffer_mgmt_config.tlsf_cfg.fl_lowest_cls = 6;
    buffer_mgmt_config.tlsf_cfg.fl_cls_cnt = 10;
    buffer_mgmt_config.tlsf_cfg.sl_cls_cnt =2;

    // init network stack
    network_init(&buffer_mgmt_config);


    uint8_t var =0;
    if(network_set_option(0xFFFF, NETWORK_OPTION_LOGGING, &var, sizeof(var)) != NET_ERR_OK)
    {
        utils_log("Error in netowrk interface when setting option NETWORK_OPTION_LOGGING\n");
        return -1;
    }

    var =0;
    if(socket_set_option(GLOBAL_SOCKET, SOCKET_OPTION_DEBUG_LOG, &var, sizeof(var) ) != NET_ERR_OK)
    {
        utils_log("Error in socket when setting option SOCKET_OPTION_DEBUG_LOG\n");
        return -1;
    }
    var =0;
    if(socket_set_option(GLOBAL_SOCKET, SOCKET_OPTION_MISSING_L2_ADDRESS_NO_FAIL, &var, sizeof(var) ) != NET_ERR_OK)
    {
        utils_log("Error in socket when setting option SOCKET_OPTION_DEBUG_LOG\n");
        return -1;
    }

    return 0;
}

void init_eth0(void)
{
    network_interface_config_t if_config;
    
    if_config.interface_name = (char*) &eth0_if_name;
    if_config.interface_type = NETWORK_INTERFACE_PATMOS_ETHERNET_TSN;

    if_config.device_config.timer_cfg.interupt_ena=1;
    if_config.device_config.timer_cfg.interupt_id = 24; // interrupts start with exception ID 16, Timer use Intterupt = 11 => exception ID 27
    if_config.device_config.timer_cfg.timer_base_address = (void*) 0xf00f0000;

    if_config.device_config.access_type = DEVICE_ACCESS_MEMORY_MAPPED_IO;
    if_config.device_config.interupt_ena = 1;
    if_config.device_config.interupt_id = 22; // interrupts start with exception ID 16, EthMac Intterupt = 6 => exception ID 22
    if_config.device_config.queue_length = 100;
    if_config.device_config.rx_buffers = 5;

    if_config.device_config.buffer_mgmt_config.type = BUFFER_MGMT_TYPE_MMIO_TLSF;
    if_config.device_config.buffer_mgmt_config.start_address = (void *) (PATMOS_IO_ETH);
    if_config.device_config.buffer_mgmt_config.end_address =  (void *) (PATMOS_IO_ETH + 0xE000);
    if_config.device_config.buffer_mgmt_config.tlsf_cfg.fl_lowest_cls = 5;//4;
    if_config.device_config.buffer_mgmt_config.tlsf_cfg.fl_cls_cnt = 11;//12;
    if_config.device_config.buffer_mgmt_config.tlsf_cfg.sl_cls_cnt =2 ;

    if_config.device_config.mac_address.length = 6;
    if_config.device_config.mac_address.address[0] = eth0_if_mac[0];
    if_config.device_config.mac_address.address[1] = eth0_if_mac[1];
    if_config.device_config.mac_address.address[2] = eth0_if_mac[2];
    if_config.device_config.mac_address.address[3] = eth0_if_mac[3];
    if_config.device_config.mac_address.address[4] = eth0_if_mac[4];
    if_config.device_config.mac_address.address[5] = eth0_if_mac[5];
    
    if_config.device_config.mmio_config.start_register_address = (void *) (PATMOS_IO_ETH + 0xF000);
    
    // layer 2 config
    if_config.layer2_config.l2_type = NETWORK_L2_ETHERNET_VLAN;
    if_config.layer2_config.vlan.dei=0;
    if_config.layer2_config.vlan.pcp=0;
    if_config.layer2_config.vlan.vid=100;

    // network address configuration
    if_config.address.type = NETWORK_ADDRESS_IP;
    if_config.address.ip.ip_address.version = IP_VERSION_4;
    if_config.address.ip.ip_address.ipv4_addr[0] = eth0_if_ipv4[0];
    if_config.address.ip.ip_address.ipv4_addr[1] = eth0_if_ipv4[1];
    if_config.address.ip.ip_address.ipv4_addr[2] = eth0_if_ipv4[2];
    if_config.address.ip.ip_address.ipv4_addr[3] = eth0_if_ipv4[3];
    if_config.address.ip.netmask = 24;

    eth0_interface_id = network_register_interface(&if_config);
    if(eth0_interface_id == 0xFFFF)
    {
        utils_log("Error in interface registration\n");
    }
}

int populate_tsn_schedule(void)
{
    tsn_gcl_config_t tsn_config;
    tsn_gcl_entry_t gcl[10];
    
    tsn_config.entry_list = (tsn_gcl_entry_t*) &(gcl);
    gcl[0].gate_state=0x7F;
    gcl[0].slot_duration= 1000000; //ns

    gcl[1].gate_state=0x00;
    gcl[1].slot_duration= 1000000; // ns

    gcl[2].gate_state=0x80;
    gcl[2].slot_duration= 1000000; // ns

    gcl[3].gate_state=0x00;
    gcl[3].slot_duration= 1000000; // ns

    tsn_config.gcl_length = 4;
    tsn_config.cycle_duration = 50000000; //1000000*(tsn_config.gcl_length+1); //ns
    
    tsn_config.admin_cycle_time.seconds=0;
    tsn_config.admin_cycle_time.nanoseconds=0;

    tsn_config.min_guard_band_duration = 1000000; //ns

    network_error_t ret = network_set_option(eth0_interface_id, NETWORK_OPTION_TSN_CFG, &tsn_config, sizeof(tsn_config));
    if(ret != NET_ERR_OK)
    {
        utils_log("network_set_option NETWORK_OPTION_TSN_CFG: 0x%04hx\n", ret);
        LEDS = 0x55;
        while(1){}
        return -1;
    }
    return 0;
}

void init_socket_eth0()
{
    // Setup main loop socket
    eth0_socket_id = socket(SOCKET_DOMAIN_IP, SOCKET_TYPE_DATAGRAM, 0);
    if(eth0_socket_id == 0xFFFF)
    {
        utils_log("Error in socket creation\n");
        return;
    }

    socket_address_t src_addr;
    src_addr.ip_address.version = IP_VERSION_4;
    src_addr.ip_address.ipv4_addr[0] = eth0_if_ipv4[0];
    src_addr.ip_address.ipv4_addr[1] = eth0_if_ipv4[1];
    src_addr.ip_address.ipv4_addr[2] = eth0_if_ipv4[2];
    src_addr.ip_address.ipv4_addr[3] = eth0_if_ipv4[3];
    src_addr.port = src_port;

    socket_bind(eth0_socket_id, &src_addr);

    uint8_t value = 0;
    network_error_t ret= socket_set_option(eth0_socket_id, SOCKET_OPTION_PRIORITY, &value, sizeof(value));
    if(ret != NET_ERR_OK)
    {
        utils_log("Error in setting socket priority\n");
        return;
    }
    
    // Setup priority socket
    eth0_isr_socket_id = socket(SOCKET_DOMAIN_IP, SOCKET_TYPE_DATAGRAM, 0);
    if(eth0_socket_id == 0xFFFF)
    {
        utils_log("Error in socket creation\n");
        return;
    }

    src_addr.port = src_port_isr;

    socket_bind(eth0_isr_socket_id, &src_addr);

    value = 7;
    ret= socket_set_option(eth0_isr_socket_id, SOCKET_OPTION_PRIORITY, &value, sizeof(value));
    if(ret != NET_ERR_OK)
    {
        utils_log("Error in setting socket priority\n");
        return;
    }
    
}

void init_dest_addr(void)
{
    /*
    dest_addr.hw_address.length = 6;
    dest_addr.hw_address.address[0] = dst_mac[0];
    dest_addr.hw_address.address[1] = dst_mac[1];
    dest_addr.hw_address.address[2] = dst_mac[2];
    dest_addr.hw_address.address[3] = dst_mac[3];
    dest_addr.hw_address.address[4] = dst_mac[4];
    dest_addr.hw_address.address[5] = dst_mac[5];
    */
    eth0_dest_addr.ip_address.version = IP_VERSION_4;
    eth0_dest_addr.ip_address.ipv4_addr[0] = eth0_dst_ipv4[0];
    eth0_dest_addr.ip_address.ipv4_addr[1] = eth0_dst_ipv4[1];
    eth0_dest_addr.ip_address.ipv4_addr[2] = eth0_dst_ipv4[2];
    eth0_dest_addr.ip_address.ipv4_addr[3] = eth0_dst_ipv4[3];
    eth0_dest_addr.port = dst_port;
}

// a basic handler for faults
void fault_handler(void) {
  unsigned source = exc_get_source();
  ////LEDS = source;

  switch(source) {
  case 0: puts("Illegal operation"); break;
  case 1: puts("Illegal memory access"); break;
  default:puts("FAULT"); break;
  }

  // cannot recover from a fault
  abort();
}

void isr_routine(void)
{
    SEI();
    static uint8_t cnt=0;
    cnt = (cnt+1) >=10 ? 0 : cnt+1;
    
    if(cnt==0)
    {
        uint8_t*data = (uint8_t*) &hello_short;
        uint16_t data_len =  strlen( (char*) &hello_short);
        network_error_t ret = socket_send(eth0_isr_socket_id, &eth0_dest_addr, data, data_len );
        
        if( ret != NET_ERR_OK && ret != NET_ERR_ARP_UNKNOWN_ADDR && 
            ret != NET_ERR_DRVR_FRAME_QUEUED && ret != NET_ERR_DRVR_TIMESLOT_TO_SHORT && ret != NET_ERR_DRVR_TSN_QUEUES_CLOSED)
        {
            utils_log("eth0 ISR socket_send: 0x%04hx\n", ret);
            if(ret != NET_ERR_ARP_UNKNOWN_ADDR)
            {
                CLI();
                while (1){}
            }
        }
        else
        {
            packets_sent++;
        }
    }
    else if(cnt == 5)
    {
        network_update();
    }
    else if(cnt == 7)
    {
        network_read_interface(eth0_interface_id); 
    }
    CLI();
}

void set_timer(void)
{
    uint8_t cnt = 0;
    do{
        next_tick_timestamp +=TICK_INTERVALL_US;
        cnt++;
    }
    while (next_tick_timestamp <= get_cpu_usecs());
    
    if(cnt > 1)
    {

#ifndef WCET_ANALYSIS
        puts("INTERRUPT WAS TOO LONG BLOCKED!!!");
#endif

        do{
            next_tick_timestamp +=TICK_INTERVALL_US;
            cnt++;
        }
        while (next_tick_timestamp <= get_cpu_usecs());
    }
    // rearm RTC Clock interrupt
    arm_usec_timer(next_tick_timestamp);
}

/**
 * @brief Handle timer tick
 *  
 * Interrupt callback function to handle timer ticks
 */

void handle_tick(void) __attribute__((naked));
void handle_tick()
{
    exc_prologue();
    
   
    isr_routine();

    set_timer();

    exc_epilogue();
}



int main(void)
{
    socket_address_t msg_src_addr;
    uint16_t data_length;
    
    char eth0_socket_buffer[200];

    uint32_t cnt=1;

    LEDS = 0x0;

    for (unsigned i = 0; i < 32; i++) {
        exc_register(i, &fault_handler);
    }

    if(init_network_and_socket() != 0)
    {
        return -1;
    }

    init_eth0();    
    if(eth0_interface_id == 0xFFFF)
    {
        return -1;
    }

    init_socket_eth0();
    if(eth0_socket_id == 0xFFFF)
    {
        return -1;
    }

#ifdef WCET_ANALYSIS
    handle_tick();
#endif

    init_dest_addr();
    
    if(populate_tsn_schedule() != 0)
    {
        return -1;
    }

    uint8_t*data = (uint8_t*) &hello_long;
    uint16_t data_len =  strlen( (char*) &hello_long);
    network_error_t ret;
    

    // configure timer interrupt
    exc_register(17, &handle_tick); 

    intr_unmask_all();
    intr_clear_all_pending();
    // enable interrupts
    intr_enable();
    
    next_tick_timestamp =get_cpu_usecs() + TICK_INTERVALL_US;
    arm_usec_timer(next_tick_timestamp);

    uint32_t packets_sent_last = 0;
    printf("hello long length: %u\n", data_len);
    //LEDS = 0x0;
    packets_received =0;
    packets_sent =0;
    while (1)
    {
        unsigned long long int start_time = get_cpu_usecs();
        while (get_cpu_usecs()-start_time < 100000){}

        //LEDS ^= 0x1; data_len
        ret = socket_send(eth0_socket_id, &eth0_dest_addr, data, data_len );
        if( ret != NET_ERR_OK && ret != NET_ERR_ARP_UNKNOWN_ADDR && 
            ret != NET_ERR_DRVR_FRAME_QUEUED && ret != NET_ERR_DRVR_TIMESLOT_TO_SHORT && ret != NET_ERR_DRVR_TSN_QUEUES_CLOSED)
        {
            utils_log("eth0 socket_send: 0x%04hx\n", ret);
            if(ret != NET_ERR_ARP_UNKNOWN_ADDR)
            {
                CLI();
                while (1){}
            }
        }
        else
        {
            packets_sent++;
        }

        //LEDS ^= 0x2;
        network_read_interface(eth0_interface_id); 
        //LEDS ^= 0x4;
        cnt =0;
        do
        {
            ret = socket_receive(eth0_socket_id, &msg_src_addr, (uint8_t*) &eth0_socket_buffer, &data_length, 200);
            if(ret != NET_ERR_SOCKET_NO_MSG && ret != NET_ERR_OK)
            {
                utils_log("eth0 socket_receive: 0x%04hx\n", ret);
            }

            if(ret == NET_ERR_OK)
            {
                cnt++;
            }
        }while (ret == NET_ERR_OK);       

        //LEDS ^= 0x8;
        do
        {
            ret = socket_receive(eth0_isr_socket_id, &msg_src_addr, (uint8_t*) &eth0_socket_buffer, &data_length, 200);
            if(ret != NET_ERR_SOCKET_NO_MSG && ret != NET_ERR_OK)
            {
                utils_log("eth0_ISR socket_receive: 0x%04hx\n", ret);
            }

            if(ret == NET_ERR_OK)
            {
                cnt++;
            }
        }while (ret == NET_ERR_OK);       
        
        //LEDS ^= 0x10;
        packets_received += cnt;

        ////LEDS = packets_sent - packets_received;
        packets_sent_last = packets_sent;
    }

    return 0;
}