#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "../inc/buffer_management.h"
#include "../inc/device.h"
#include "../inc/generic_types.h"
#include "../inc/net_types.h"
#include "../inc/network.h"
#include "../inc/socket.h"
#include "../inc/utils.h"
#include "../library/driver/ethernet/patmos_eth_mac.h"

#include <machine/exceptions.h>
#include <machine/patmos.h>
#include <machine/rtc.h>
#include <machine/uart.h>

// Network Stack config
#ifndef NETWORK_STACK_MEMORY_POOL_SIZE
    #define NETWORK_STACK_MEMORY_POOL_SIZE 0xE000
#endif
uint8_t network_stack_buffer[NETWORK_STACK_MEMORY_POOL_SIZE];

// Network interface config
char eth0_if_name[] = "eth0";
uint8_t eth0_if_ipv4[4] = {172, 16, 0, 201};
uint8_t eth0_if_mac[6] = {0x00, 0x80, 0x6e, 0xF0, 0xDA, 0x42};
uint16_t eth0_interface_id= 0xFFFF;

char eth1_if_name[] = "eth1";
uint8_t eth1_if_ipv4[4] = {172, 16, 1, 201};
uint8_t eth1_if_mac[6] = {0x00, 0x80, 0x6e, 0xF0, 0xDA, 0x43};
uint16_t eth1_interface_id= 0xFFFF;

// socket configuration
uint16_t eth0_socket_id= 0xFFFF;
uint16_t eth1_socket_id= 0xFFFF;
unsigned short src_port = 54321;


// destination config
socket_address_t eth0_dest_addr;
socket_address_t eth1_dest_addr;
//unsigned char dst_mac[6] ={0x00, 0x0e, 0xc6, 0xc9, 0x48, 0x59};
unsigned char eth0_dst_ipv4[4] = {172, 16, 0, 200};
unsigned char eth1_dst_ipv4[4] = {172, 16, 1, 200};
unsigned short dst_port = 12345;
char hello_short[] = "hello ISR";
char hello_long[] = "hello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!\thello world!"; //\thello world!";

// Timer interrupt
unsigned long long next_tick_timestamp;
volatile unsigned long long int isr_runtime;
unsigned long long int isr_runtime_longest;
unsigned long long int isr_start_delay;
unsigned long long int isr_start_delay_longest;

volatile uint32_t packets_sent = 0;
volatile uint32_t packets_received = 0;

volatile uint8_t isr_fired=0;
#define TICK_INTERVALL_US 10000
#define LEDS (*((volatile _IODEV unsigned *)  PATMOS_IO_LED))

int init_network_and_socket(void)
{
    buffer_management_config_t buffer_mgmt_config;
    buffer_mgmt_config.type = BUFFER_MGMT_TYPE_TLSF;
    buffer_mgmt_config.start_address = (void *) (network_stack_buffer);
    buffer_mgmt_config.end_address =  (void *) (network_stack_buffer) + NETWORK_STACK_MEMORY_POOL_SIZE;
    buffer_mgmt_config.tlsf_cfg.fl_lowest_cls = 6;
    buffer_mgmt_config.tlsf_cfg.fl_cls_cnt = 10;
    buffer_mgmt_config.tlsf_cfg.sl_cls_cnt =2;

    // init network stack
    network_init(&buffer_mgmt_config);


    uint8_t var =0;
    if(network_set_option(0xFFFF, NETWORK_OPTION_LOGGING, &var, sizeof(var)) != NET_ERR_OK)
    {
        utils_log("Error in netowrk interface when setting option NETWORK_OPTION_LOGGING\n");
        return -1;
    }

    var =0;
    if(socket_set_option(GLOBAL_SOCKET, SOCKET_OPTION_DEBUG_LOG, &var, sizeof(var) ) != NET_ERR_OK)
    {
        utils_log("Error in socket when setting option SOCKET_OPTION_DEBUG_LOG\n");
        return -1;
    }
    var =0;
    if(socket_set_option(GLOBAL_SOCKET, SOCKET_OPTION_MISSING_L2_ADDRESS_NO_FAIL, &var, sizeof(var) ) != NET_ERR_OK)
    {
        utils_log("Error in socket when setting option SOCKET_OPTION_DEBUG_LOG\n");
        return -1;
    }

    return 0;
}

void init_eth0(void)
{
    network_interface_config_t if_config;
    
    if_config.interface_name = (char*) &eth0_if_name;
    if_config.interface_type = NETWORK_INTERFACE_PATMOS_ETHERNET;

    if_config.device_config.access_type = DEVICE_ACCESS_MEMORY_MAPPED_IO;
    if_config.device_config.interupt_ena = 1;
    if_config.device_config.interupt_id = 22; // interrupts start with exception ID 16, EthMac Intterupt = 6 => exception ID 22
    if_config.device_config.queue_length = 100;
    if_config.device_config.rx_buffers = 5;

    if_config.device_config.buffer_mgmt_config.type = BUFFER_MGMT_TYPE_MMIO_TLSF;
    if_config.device_config.buffer_mgmt_config.start_address = (void *) (PATMOS_IO_ETH);
    if_config.device_config.buffer_mgmt_config.end_address =  (void *) (PATMOS_IO_ETH + 0xE000);
    if_config.device_config.buffer_mgmt_config.tlsf_cfg.fl_lowest_cls = 5;//4;
    if_config.device_config.buffer_mgmt_config.tlsf_cfg.fl_cls_cnt = 11;//12;
    if_config.device_config.buffer_mgmt_config.tlsf_cfg.sl_cls_cnt =2 ;

    if_config.device_config.mac_address.length = 6;
    if_config.device_config.mac_address.address[0] = eth0_if_mac[0];
    if_config.device_config.mac_address.address[1] = eth0_if_mac[1];
    if_config.device_config.mac_address.address[2] = eth0_if_mac[2];
    if_config.device_config.mac_address.address[3] = eth0_if_mac[3];
    if_config.device_config.mac_address.address[4] = eth0_if_mac[4];
    if_config.device_config.mac_address.address[5] = eth0_if_mac[5];
    
    if_config.device_config.mmio_config.start_register_address = (void *) (PATMOS_IO_ETH + 0xF000);
    // layer 2 config
    if_config.layer2_config.l2_type = NETWORK_L2_ETHERNET;

    // network address configuration
    if_config.address.type = NETWORK_ADDRESS_IP;
    if_config.address.ip.ip_address.version = IP_VERSION_4;
    if_config.address.ip.ip_address.ipv4_addr[0] = eth0_if_ipv4[0];
    if_config.address.ip.ip_address.ipv4_addr[1] = eth0_if_ipv4[1];
    if_config.address.ip.ip_address.ipv4_addr[2] = eth0_if_ipv4[2];
    if_config.address.ip.ip_address.ipv4_addr[3] = eth0_if_ipv4[3];
    if_config.address.ip.netmask = 24;

    eth0_interface_id = network_register_interface(&if_config);
    if(eth0_interface_id == 0xFFFF)
    {
        utils_log("Error in interface registration\n");
    }
}

void init_socket_eth0()
{

    eth0_socket_id = socket(SOCKET_DOMAIN_IP, SOCKET_TYPE_DATAGRAM, 0);
    if(eth0_socket_id == 0xFFFF)
    {
        utils_log("Error in socket creation\n");
        return;
    }

    socket_address_t src_addr;
    src_addr.ip_address.version = IP_VERSION_4;
    src_addr.ip_address.ipv4_addr[0] = eth0_if_ipv4[0];
    src_addr.ip_address.ipv4_addr[1] = eth0_if_ipv4[1];
    src_addr.ip_address.ipv4_addr[2] = eth0_if_ipv4[2];
    src_addr.ip_address.ipv4_addr[3] = eth0_if_ipv4[3];
    src_addr.port = src_port;

    socket_bind(eth0_socket_id, &src_addr);
}

#ifdef PATMOS_UDP_HELLO_DUAL_ETHERNET
void init_eth1(void)
{
    network_interface_config_t if_config;
    
    if_config.interface_name = (char*) &eth1_if_name;
    if_config.interface_type = NETWORK_INTERFACE_PATMOS_ETHERNET;

    if_config.device_config.access_type = DEVICE_ACCESS_MEMORY_MAPPED_IO;
    if_config.device_config.interupt_ena = 1;
    if_config.device_config.interupt_id = 25; // interrupts start with exception ID 16, EthMac 1 Intterupt = 9 => exception ID 25
    if_config.device_config.queue_length = 100;
    if_config.device_config.rx_buffers = 5;

    if_config.device_config.buffer_mgmt_config.type = BUFFER_MGMT_TYPE_MMIO_TLSF;
    if_config.device_config.buffer_mgmt_config.start_address = (void *) (PATMOS_IO_ETH1);
    if_config.device_config.buffer_mgmt_config.end_address =  (void *) (PATMOS_IO_ETH1 + 0xE000);
    if_config.device_config.buffer_mgmt_config.tlsf_cfg.fl_lowest_cls = 5;//4;
    if_config.device_config.buffer_mgmt_config.tlsf_cfg.fl_cls_cnt = 11;//12;
    if_config.device_config.buffer_mgmt_config.tlsf_cfg.sl_cls_cnt =2 ;

    if_config.device_config.mac_address.length = 6;
    if_config.device_config.mac_address.address[0] = eth1_if_mac[0];
    if_config.device_config.mac_address.address[1] = eth1_if_mac[1];
    if_config.device_config.mac_address.address[2] = eth1_if_mac[2];
    if_config.device_config.mac_address.address[3] = eth1_if_mac[3];
    if_config.device_config.mac_address.address[4] = eth1_if_mac[4];
    if_config.device_config.mac_address.address[5] = eth1_if_mac[5];
    
    if_config.device_config.mmio_config.start_register_address = (void *) (PATMOS_IO_ETH1 + 0xF000);
    // layer 2 config
    if_config.layer2_config.l2_type = NETWORK_L2_ETHERNET;

    // network address configuration
    if_config.address.type = NETWORK_ADDRESS_IP;
    if_config.address.ip.ip_address.version = IP_VERSION_4;
    if_config.address.ip.ip_address.ipv4_addr[0] = eth1_if_ipv4[0];
    if_config.address.ip.ip_address.ipv4_addr[1] = eth1_if_ipv4[1];
    if_config.address.ip.ip_address.ipv4_addr[2] = eth1_if_ipv4[2];
    if_config.address.ip.ip_address.ipv4_addr[3] = eth1_if_ipv4[3];
    if_config.address.ip.netmask = 24;

    eth1_interface_id = network_register_interface(&if_config);
    if(eth1_interface_id == 0xFFFF)
    {
        utils_log("Error in interface registration\n");
    }
}

void init_socket_eth1()
{

    eth1_socket_id = socket(SOCKET_DOMAIN_IP, SOCKET_TYPE_DATAGRAM, 0);
    if(eth1_socket_id == 0xFFFF)
    {
        utils_log("Error in socket creation\n");
        return;
    }

    socket_address_t src_addr;
    src_addr.ip_address.version = IP_VERSION_4;
    src_addr.ip_address.ipv4_addr[0] = eth1_if_ipv4[0];
    src_addr.ip_address.ipv4_addr[1] = eth1_if_ipv4[1];
    src_addr.ip_address.ipv4_addr[2] = eth1_if_ipv4[2];
    src_addr.ip_address.ipv4_addr[3] = eth1_if_ipv4[3];
    src_addr.port = src_port;

    socket_bind(eth1_socket_id, &src_addr);
}

#endif

void init_dest_addr(void)
{
    /*
    dest_addr.hw_address.length = 6;
    dest_addr.hw_address.address[0] = dst_mac[0];
    dest_addr.hw_address.address[1] = dst_mac[1];
    dest_addr.hw_address.address[2] = dst_mac[2];
    dest_addr.hw_address.address[3] = dst_mac[3];
    dest_addr.hw_address.address[4] = dst_mac[4];
    dest_addr.hw_address.address[5] = dst_mac[5];
    */
    eth0_dest_addr.ip_address.version = IP_VERSION_4;
    eth0_dest_addr.ip_address.ipv4_addr[0] = eth0_dst_ipv4[0];
    eth0_dest_addr.ip_address.ipv4_addr[1] = eth0_dst_ipv4[1];
    eth0_dest_addr.ip_address.ipv4_addr[2] = eth0_dst_ipv4[2];
    eth0_dest_addr.ip_address.ipv4_addr[3] = eth0_dst_ipv4[3];
    eth0_dest_addr.port = dst_port;

    eth1_dest_addr.ip_address.version = IP_VERSION_4;
    eth1_dest_addr.ip_address.ipv4_addr[0] = eth1_dst_ipv4[0];
    eth1_dest_addr.ip_address.ipv4_addr[1] = eth1_dst_ipv4[1];
    eth1_dest_addr.ip_address.ipv4_addr[2] = eth1_dst_ipv4[2];
    eth1_dest_addr.ip_address.ipv4_addr[3] = eth1_dst_ipv4[3];
    eth1_dest_addr.port = dst_port;
}

// a basic handler for faults
void fault_handler(void) {
  unsigned source = exc_get_source();
  ////LEDS = source;

  switch(source) {
  case 0: puts("Illegal operation"); break;
  case 1: puts("Illegal memory access"); break;
  default:puts("FAULT"); break;
  }

  // cannot recover from a fault
  abort();
}

void isr_routine(void)
{
    //utils_log("in isr_routine\n");
    SEI();
    static uint8_t cnt=0;
    cnt = (cnt+1) >=10 ? 0 : cnt+1;
    
    if(cnt==0)
    {
        uint8_t*data = (uint8_t*) &hello_short;
        uint16_t data_len =  strlen( (char*) &hello_short);
        network_error_t ret = socket_send(eth0_socket_id, &eth0_dest_addr, data, data_len );
        
        if( ret != NET_ERR_OK && ret != NET_ERR_ARP_UNKNOWN_ADDR && 
            ret != NET_ERR_DRVR_FRAME_QUEUED && ret != NET_ERR_DRVR_TIMESLOT_TO_SHORT && ret != NET_ERR_DRVR_TSN_QUEUES_CLOSED)
        {
            utils_log("eth0 ISR socket_send: 0x%04hx\n", ret);
            if(ret != NET_ERR_ARP_UNKNOWN_ADDR)
            {
                CLI();
                while (1){}
            }
        }
        else
        {
            packets_sent++;
        }
    }
#ifdef PATMOS_UDP_HELLO_DUAL_ETHERNET
    else if(cnt == 2)
    {
        uint8_t*data = (uint8_t*) &hello_short;
        uint16_t data_len =  strlen( (char*) &hello_short);
        
        network_error_t ret = socket_send(eth1_socket_id, &eth1_dest_addr, data, data_len );
        if(ret == NET_ERR_OK || ret == NET_ERR_DRVR_FRAME_QUEUED)
        {
            packets_sent++;
        }
        else if(ret != NET_ERR_ARP_UNKNOWN_ADDR)
        {
            utils_log("eth1 socket_send: 0x%04hx\n", ret);
            while (1){}
        }
    }
#endif
    else if(cnt == 5)
    {
        network_update();
    }
    else if(cnt == 7)
    {
        network_read_interface(eth0_interface_id); 
    }
#ifdef PATMOS_UDP_HELLO_DUAL_ETHERNET
    else if (cnt == 9)
    {
        network_read_interface(eth1_interface_id);
    }
#endif

    CLI();
}

void set_timer(void)
{
    uint8_t cnt = 0;
    do{
        next_tick_timestamp +=TICK_INTERVALL_US;
        cnt++;
    }
    while (next_tick_timestamp <= get_cpu_usecs());
    
    if(cnt > 1)
    {

#ifndef WCET_ANALYSIS
        puts("INTERRUPT WAS TOO LONG BLOCKED!!!");
#endif

        do{
            next_tick_timestamp +=TICK_INTERVALL_US;
            cnt++;
        }
        while (next_tick_timestamp <= get_cpu_usecs());
    }
    // rearm RTC Clock interrupt
    arm_usec_timer(next_tick_timestamp);
}

/**
 * @brief Handle timer tick
 *  
 * Interrupt callback function to handle timer ticks
 */

void handle_tick(void) __attribute__((naked));
void handle_tick()
{
    exc_prologue();
    
   
    isr_routine();

    set_timer();

    exc_epilogue();
}



int main(void)
{
    socket_address_t msg_src_addr;
    uint16_t data_length;
    
    char eth0_socket_buffer[200];
    socket_buffer_cfg_t eth0_socket_rcv_cfg;
    eth0_socket_rcv_cfg.buffer_entry_size = 200;
    eth0_socket_rcv_cfg.buffer_size = eth0_socket_rcv_cfg.buffer_entry_size *40;
    uint8_t eth0_socket_rcv_buffer_space[eth0_socket_rcv_cfg.buffer_size];
    eth0_socket_rcv_cfg.buffer_space = (uint8_t*) &eth0_socket_rcv_buffer_space;

    uint32_t cnt=1;

    LEDS = 0x0;

    for (unsigned i = 0; i < 32; i++) {
        exc_register(i, &fault_handler);
    }

    if(init_network_and_socket() != 0)
    {
        return -1;
    }

    init_eth0();    
    if(eth0_interface_id == 0xFFFF)
    {
        return -1;
    }

    init_socket_eth0();
    if(eth0_socket_id == 0xFFFF)
    {
        return -1;
    }

#ifdef WCET_ANALYSIS
    handle_tick();
#endif

#ifdef PATMOS_UDP_HELLO_DUAL_ETHERNET    

    char eth1_socket_buffer[200];

    init_eth1();    
    if(eth1_interface_id == 0xFFFF)
    {
        return -1;
    }

    init_socket_eth1();
    if(eth1_socket_id == 0xFFFF)
    {
        return -1;
    }
#endif

    init_dest_addr();
    
    uint8_t*data = (uint8_t*) &hello_long;
    uint16_t data_len =  strlen( (char*) &hello_long);
    network_error_t ret;

    // configure timer interrupt
    exc_register(17, &handle_tick); 

    intr_unmask_all();
    intr_clear_all_pending();
    // enable interrupts
    intr_enable();
    
    next_tick_timestamp =get_cpu_usecs() + TICK_INTERVALL_US;
    arm_usec_timer(next_tick_timestamp);

    uint32_t packets_sent_last = 0;

    int passWhile = 10;
    LEDS = 0x0;
    packets_received =0;
    packets_sent =0;
    while (1)   // to test : while(passWhile > 0)
    {
        // utils_log("new while, passwhile = %d\n", passWhile);
        unsigned long long int start_time = get_cpu_usecs();
        while (get_cpu_usecs()-start_time < 1000000){}

        ret = socket_send(eth0_socket_id, &eth0_dest_addr, data, data_len );
        if( ret != NET_ERR_OK && ret != NET_ERR_ARP_UNKNOWN_ADDR && 
            ret != NET_ERR_DRVR_FRAME_QUEUED && ret != NET_ERR_DRVR_TIMESLOT_TO_SHORT && ret != NET_ERR_DRVR_TSN_QUEUES_CLOSED)
        {
            utils_log("eth0 socket_send: 0x%04hx\n", ret);
            if(ret != NET_ERR_ARP_UNKNOWN_ADDR)
            {
                CLI();
                utils_log("ret != NET_ERR_ARP_UNKNOWN_ADDR and NET_ERR_OK and ...\n");
                utils_log("while(1)\n");
                while (1){}
            }
        }
        else
        {
            packets_sent++;
        }
        
#ifdef PATMOS_UDP_HELLO_DUAL_ETHERNET
        ret = socket_send(eth1_socket_id, &eth1_dest_addr, data, data_len );
        if(ret != NET_ERR_OK && ret != NET_ERR_DRVR_FRAME_QUEUED)
        {
            utils_log("eth1 socket_send: 0x%04hx\n", ret);
            if(ret != NET_ERR_ARP_UNKNOWN_ADDR)
            {
                while (1){}
            }
        }
        else
        {
            packets_sent++;
        }
#endif

        cnt =0;
        do
        {
            ret = socket_receive(eth0_socket_id, &msg_src_addr, (uint8_t*) &eth0_socket_buffer, &data_length, 200);
            if(ret != NET_ERR_SOCKET_NO_MSG && ret != NET_ERR_OK)
            {
                utils_log("eth0 socket_receive: 0x%04hx\n", ret);
            }

            if(ret == NET_ERR_OK)
            {
                cnt++;
            }
        }while (ret == NET_ERR_OK);       
        
#ifdef PATMOS_UDP_HELLO_DUAL_ETHERNET
        do
        {
            ret = socket_receive(eth1_socket_id, &msg_src_addr, (uint8_t*) &eth1_socket_buffer, &data_length, 200);
            if(ret != NET_ERR_SOCKET_NO_MSG && ret != NET_ERR_OK)
            {
                utils_log("eth1 socket_receive: 0x%04hx\n", ret);
            }

            if(ret == NET_ERR_OK)
            {
                cnt++;
            }
        }while (ret == NET_ERR_OK);     
#endif

        packets_received += cnt;

        LEDS = packets_sent - packets_received;
        packets_sent_last = packets_sent;
        passWhile--;
    }

    utils_log("\nend prog\n");
    return 0;
}