
#ifndef _INCLUDE_BUFFER_MANAGEMENT_H_
#define _INCLUDE_BUFFER_MANAGEMENT_H_

#ifndef MAX_BUFFER_CONFIGS
    #define MAX_BUFFER_CONFIGS 32
#endif

#include "generic_types.h"

typedef enum
{
   BUFFER_MGMT_TYPE_NONE            = 0x00, 
   BUFFER_MGMT_TYPE_UNIFORM         = 0x01,
   BUFFER_MGMT_TYPE_TLSF            = 0x02,
   BUFFER_MGMT_TYPE_MMIO_UNIFORM    = 0x81,
   BUFFER_MGMT_TYPE_MMIO_TLSF       = 0x82,
} buffer_mgmt_type_t;

typedef struct 
{
    unsigned short buffer_size;
} uniform_config_t;


typedef struct 
{
	unsigned char fl_lowest_cls;
	unsigned char fl_cls_cnt;
	unsigned char sl_cls_cnt;
} tlsf_config_t;

typedef struct 
{
    buffer_mgmt_type_t type;
    void * start_address;
    void * end_address;
    union 
    {
        tlsf_config_t tlsf_cfg;
        uniform_config_t uniform_cfg;
    };
   
} buffer_management_config_t;

// register a new memory space that shall be managed
unsigned short buffer_management_register(buffer_management_config_t* config );
// remove a memory space that is currently managed
void buffer_management_derigister(unsigned short buf_mgmt_id);
// allocate a buffer 
void * buffer_management_malloc(unsigned short size, unsigned short buf_mgmt_id);
// reallocate a buffer by splitting it into two parts, if splitting is not possible NULL is returned
void * buffer_management_realloc(void *  buffer_in, unsigned short size, unsigned short buf_mgmt_id);

// freea a buffer again
void buffer_management_free(void * buffer, unsigned short buf_mgmt_id);

void buffer_management_memory_dump(unsigned short buf_mgmt_id);

#endif
