#ifndef DRIVER_DEVICE_H_
#define DRIVER_DEVICE_H_

#include "generic_types.h"
#include "net_types.h"
#include "buffer_management.h"

#ifndef DEVICE_MAX_RX_BUFFERS
    #define DEVICE_MAX_RX_BUFFERS 5
#endif

typedef enum
{
    DEVICE_ACCESS_MEMORY_MAPPED_IO  = 0x00,
    DEVICE_ACCESS_SPI               = 0x01,
    DEVICE_ACCESS_UART              = 0x02,
} device_access_type_t;

typedef struct 
{
    void* start_register_address;
} device_mmio_config_t;

typedef struct
{
    uint64_t gate_state      :8;
    uint64_t slot_duration  :56;
} tsn_gcl_entry_t;

typedef struct 
{
    uint64_t cycle_duration;            // full cycle duration
    time_stamp_t admin_cycle_time;          // time when configuration shall be switched   
    uint64_t min_guard_band_duration;   // minimum duration of guard band
    tsn_gcl_entry_t *entry_list;
    uint16_t gcl_length; 
} tsn_gcl_config_t;

typedef struct 
{
    uint16_t interupt_id:15; 
    uint16_t interupt_ena:1;
    void * timer_base_address;
} timer_cfg_t;

typedef struct 
{
    device_access_type_t access_type;
    uint8_t queue_count;
    // interrupt configuration
    uint16_t interupt_id:15;
    uint16_t interupt_ena:1;
    timer_cfg_t timer_cfg; // optional
    buffer_management_config_t buffer_mgmt_config;  // configuration for buffer management
    mac_address_t mac_address;                      // ignored if not configureable
    union 
    {
        device_mmio_config_t mmio_config;
    };
    uint16_t queue_length;
    uint16_t rx_buffers;
} device_config_t;

#endif
