
#ifndef LIBNETWORK_NET_TYPES_H
#define LIBNETWORK_NET_TYPES_H

#include "generic_types.h"

/*******************************************************
 * Network Data Types 
 *******************************************************/

// Generic Error Type of library
typedef enum 
{
    NET_ERR_OK                             =  0x0000,  // No Error
    NET_ERR_BUFF_UNDERSIZED                =  0x0001,  // Buffer to small
    NET_ERR_NOT_SUPPORTED                  =  0x0002,  // Operation is not supported
    NET_ERR_INPUT_VALUE                    =  0x0003,  // an undefined input value was provided
    NET_ERR_ADDR_INCOMPATIBLE              =  0x0004,  // Network address not compatible  
    NET_ERR_TIMESTAMP_NOT_AVAILABLE        =  0x0005,  // No timestamp is is available to be consumed
    NET_ERR_DRVR_ETH                       =  0x0100,  // Generic Ethernet Driver Error
    NET_ERR_DRVR_TX_QUEUE_FULL             =  0x0101,  // Transmission Queue is full
    NET_ERR_DRVR_RX_QUEUE_FULL             =  0x0102,  // Reception Queue is full
    NET_ERR_DRVR_TX_QUEUE_EMPTY            =  0x0103,  // Transmission Queue is empty
    NET_ERR_DRVR_RX_QUEUE_EMPTY            =  0x0104,  // Reception Queue is empty
    NET_ERR_DRVR_NO_FREE_BUFFER            =  0x0105,  // No Buffer available
    NET_ERR_DRVR_FRAME_QUEUED              =  0x0106,  // Transmission is busy, frame was added to queue
    NET_ERR_DRVR_NO_FRAME_RECEIVED         =  0x0107,  // No Frame has been received
    NET_ERR_DRVR_BUFFER_BUSY               =  0x0108,  // Buffer is still being processed 
    NET_ERR_DRVR_MEM_FATAL                 =  0x0109,  // Fatal memory management error 
    NET_ERR_DRVR_MAX_BUFFER_SIZE_EXCEEDED  =  0x010A,  // Requested Buffer size exceeds maximum Buffer size
    NET_ERR_DRVR_TSN_QUEUES_CLOSED         =  0x010B,  // All TSN queues are closed
    NET_ERR_DRVR_QUEUE_FATAL               =  0x010C,  // Fatal error in queue management 
    NET_ERR_DRVR_PRIORITY                  =  0x010D,  // Priority Value not supported
    NET_ERR_DRVR_TIMESLOT_TO_SHORT         =  0x010E,  // Frame could not be sent, the active time slot is not long enough
    NET_ERR_DRVR_FRAME_TOO_LONG            =  0x010F,  // Frame execeeds maximum transmission unit of medium
    NET_ERR_ETH                            =  0x0200,  // Generic Ethernet Error
    NET_ERR_ETH_WRONG_ADDR                 =  0x0201,  // Wrong MAC Adress
    NET_ERR_ETH_NO_VLAN                    =  0x0202,  // VLAN Header expected but not found
    NET_ERR_ETH_WRONG_VLAN                 =  0x0203,  // VLAN Header has wrong VLAN ID
    NET_ERR_ETH_FRAME_LENGTH               =  0x0204,  // Ethernet Frame is short than minimum frame length
    NET_ERR_ARP                            =  0x0300,  // Generic ARP Error 
    NET_ERR_ARP_UNKNOWN_ADDR               =  0x0301,  // Requested Addres is not known
    NET_ERR_ARP_WRONG_PADDR                =  0x0302,  // Wrong Protocol Addres i.e. wrong IP
    NET_ERR_ARP_FULL                       =  0x0303,  // ARP Table is full, address could not be added
    NET_ERR_ARP_RESPONSE                   =  0x0304,  // ARP generate a response which needs to be sent
    NET_ERR_IP                             =  0x0400,  // Generic IP Error
    NET_ERR_IP_VERSION                     =  0x0401,  // Incompatible/Unknown IP Version
    NET_ERR_IP_HDR_CHCKSUM                 =  0x0402,  // Header Checksum Wrong
    NET_ERR_IP_WRONG_ADDR                  =  0x0403,  // Wrong Destination IP Address
    NET_ERR_IP_FRAGMENTED                  =  0x0404,  // IP Packet is fragment, which is not supported
    NET_ERR_IP_PACKET_LENGTH               =  0x0405,  // Total length of IP header field does not match frame length
    NET_ERR_ICMP                           =  0x0500,  // Generic ICMP Error
    NET_ERR_UDP                            =  0x0600,  // Generic UDP Error
    NET_ERR_UDP_CHCKSUM                    =  0x0601,  // UDP Checksum Error 
    NET_ERR_UDP_MSG_LENGTH                 =  0x0602,  // Incorrect UDP Message length
    NET_ERR_TCP                            =  0x0700,  // Generic TCP Error 
    NET_ERR_TCP_CHCKSUM                    =  0x0701,  // TCP Checksum Error 
    NET_ERR_TCP_MSG_LENGTH                 =  0x0702,  // Incorrect TCP Message length
    NET_ERR_TCP_NO_SERVER                  =  0x0703,  // Server TCP not set
    NET_ERR_TCP_CONNECTION_NOT_FOUND       =  0x0704,  // Can't find the socket connection in the tcp server
    NET_ERR_TCP_SAVE_HDR                   =  0x0705,  // No more place to save the header when a frame arrived
    NET_ERR_TCP_CONNECTION_STATE           =  0x0706,  // State of the connection unknow
    NET_ERR_TCP_RETRANSMISSION_QUEUE       =  0x0707,  // Problem with the retransmission queue in TCP server
    NET_ERR_PTP                            =  0x0800,  // Generic PTP Error
    NET_ERR_MGMT                           =  0x0900,  // Generic Network Managemt
    NET_ERR_MGMT_UNKNOWN                   =  0x0901,  // Unknown Network Interface or Interface Type
    NET_ERR_MGMT_NO_MSG                    =  0x0902,  // No Message received
    NET_ERR_MGMT_MEM_FULL                  =  0x0903,  // No free memory in stacks memory
    NET_ERR_SOCKET                         =  0x0A00,  // Generic Socket Error
    NET_ERR_SOCKET_UNKNOWN                 =  0x0A01,  // Unknown Socket ID given
    NET_ERR_SOCKET_PROTOCOL_NOT_CONFIG     =  0x0A02,  // No socket with matching protocol configured
    NET_ERR_SOCKET_NO_RCV_BUFFER           =  0x0A03,  // Socket has no receive buffer configured
    NET_ERR_SOCKET_MSMG_LEN_EXCEEDED       =  0x0A04,  // The received message exceeds the length of the receive buffer
    NET_ERR_SOCKET_NO_MSG                  =  0x0A05,  // No message was received
    NET_ERR_SOCKET_RCV_BUFFER_FULL         =  0x0A06,  // The receive Buffer of the socket is full
    NET_ERR_SOCKET_NOT_BOUND               =  0x0A07,  // Socket with level 2 protocol is not bound to an interface

    NET_ERR_TIME_SYNC                      =  0xF000,  // Generic Time Sync Error
    NET_ERR_TIME_SYNC_WRONG_STATE          =  0xF001,  // State violation
    NET_ERR_TIME_SYNC_WRONG_SEQUENCE_ID    =  0xF002,  // Sequenced ID violation
   
} network_error_t;

typedef enum sgmnt_type
{
    SEGMENT_TYPE_ETHERNET,
    SEGMENT_TYPE_ETHERNET_VLAN,
    SEGMENT_TYPE_IPv4,
    SEGMENT_TYPE_IPv6,
    SEGMENT_TYPE_ARP,
    SEGMENT_TYPE_ICMP,
    SEGMENT_TYPE_TCP,
    SEGMENT_TYPE_UDP,
    SEGMENT_TYPE_HEADER_DATA,
    SEGMENT_TYPE_USER_DATA,
    SEGMENT_TYPE_UNKNOWN,
} segment_type_t;


// Data type for a segment of a network frame
typedef struct network_frame
{
    uint16_t header_length;       // length of this segment
    uint16_t user_data_length;    // length of segments appended to this
    uint8_t *header_data;         // data of this segment
    uint8_t *user_data;           // next segment of the frame
    uint8_t priority;             // frame priority
} network_frame_t;

typedef struct 
{
    uint8_t length;      // MAC address can bei either EUI-48 or EUI-64
    uint8_t address[8];  // which are 48/64 bits long
} mac_address_t;

typedef enum 
{
    ETHERTYPE_IPv4       =  0x0800, 
    ETHERTYPE_ARP        =  0x0806, 
    ETHERTYPE_VLAN       =  0x8100,
    ETHERTYPE_IPv6       =  0x86DD,
    ETHERTYPE_PTP        =  0x88F7, 
    ETHERTYPE_LLDP       =  0x88CC,
    ETHERTYPE_TTE_PCF    =  0x891D,
    ETHERTYPE_REESERVED  =  0xFFFF,
} ethertype_t; 

typedef struct 
{
    uint16_t pcp:   3;
    uint16_t dei:   1; 
    uint16_t vid:  12;
} vlan_header_t;

typedef struct 
{
    vlan_header_t vlan_header;
    ethertype_t eth_type;
} ethernet_vlan_tag_t;

typedef struct 
{
    uint8_t dest_mac[6];
    uint8_t src_mac[6];
    ethertype_t eth_type;  
} ethernet_header_t;

typedef struct 
{
    uint8_t dest_mac[6];
    uint8_t src_mac[6];
    ethertype_t vlan_eth_type;  
    uint16_t pcp:   3;
    uint16_t dei:   1; 
    uint16_t vid:  12;
    ethertype_t eth_type;
} ethernet_header_vlan_t;


typedef enum arp_operation
{
    ARP_OPERATION_REQUEST   =   0x0001,     // ARP Request
    ARP_OPERATION_RESPONSE  =   0x0002,     // ARP Response
    ARP_OPERATIOM_SIZE      =   0xFFFF,     // modified to use 16 bit
} arp_operation_t;

typedef struct 
{
    uint16_t htype;       // Hardware Type
    ethertype_t ptype;          // Protocol Type
    uint8_t hlen;         // Hardware Address Length
    uint8_t plen;         // Protocol Address Length
    arp_operation_t oper;       // Operation
    uint8_t sha[6];       // Sender Hardware Address
    uint8_t spa[4];       // Sender Protocol Address
    uint8_t tha[6];       // Target Hardware Address
    uint8_t tpa[4];       // Traget Protocol Address
} arp_msg_t;

typedef enum 
{
    IP_PROTOCOL_ICMP     = 0x01,
    IP_PROTOCOL_UDP      = 0x11,
    IP_PROTOCOL_TCP      = 0x06,
    IP_PROTOCOL_RESERVED = 0xFF,
} ip_protocol_t;

typedef enum
{
    IP_VERSION_4        =  0x4,   // IPv4
    IP_VERSION_6        =  0x6,   // IPv6
} ip_version_t;

typedef struct 
{
    uint8_t version;
    union 
    {
        uint8_t  ipv4_addr[4];
        uint16_t ipv6_addr[8];
    };
} ip_address_t;

// IPV4 header
typedef union
{
    struct 
    {
        uint8_t version            :  4;    // Protocol Version
        uint8_t ihl                :  4;    // IP Header length
        uint8_t dscp               :  6;    // Differentiated Services Code Point use for ToS
        uint8_t ecn                :  2;    // Explicit Congestion Notification
        uint16_t total_length;              // Total length of header + data
        uint16_t identification;            // Identification of IP Packet for fragmentation
        uint16_t flags             :  3;    // Fragmentation Flags
        uint16_t fragment_offset   :  13;   // Offset in Bytes of this fragment
        uint8_t ttl;                        // Remaining Lifetime in hops of this packet
        ip_protocol_t protocol;             // Encapusalted protocol
        uint16_t header_checksum;           // Checksum of the header
        uint8_t source_ip[4];               // Origin of this packet
        uint8_t destination_ip[4];          // Destination of this packet
    };
    uint32_t raw[5];
} ipv4_header_t;

// UDP header
typedef union
{
    struct 
    {
        uint16_t src_port;
        uint16_t dest_port;
        uint16_t length;
        uint16_t checksum;
    };
    uint32_t raw[2];
} udp_header_t;

// TCP header
typedef union
{
    struct 
    {
        uint16_t src_port;
        uint16_t dest_port;
        uint32_t sequence_number;
        uint32_t acknowledgement_number;
        uint16_t flags;
        uint16_t window_size;
        uint16_t checksum;
        uint16_t urgent_pointer;
    };
    uint32_t raw[5];
} tcp_header_t;

// // OLD Implementation
// typedef union
// {
//     struct 
//     {
//         uint16_t src_port;
//         uint16_t dest_port;
//         uint32_t sequence_number;
//         uint32_t acknowledgement_number;
//         union
//         {
//             struct
//             {
//                 uint16_t data_offset :  4;
//                 uint16_t reserved    :  3;
//                 uint16_t ns          :  1;
//                 uint16_t cwr         :  1;
//                 uint16_t ece         :  1;
//                 uint16_t urg         :  1;
//                 uint16_t ack         :  1;
//                 uint16_t psh         :  1;
//                 uint16_t rst         :  1;
//                 uint16_t syn         :  1;
//                 uint16_t fin         :  1;
//             };
//             uint16_t raw;
//         } flags;
//         uint16_t window_size;
//         uint16_t checksum;
//         uint16_t urgent_pointer;
//     };
//     uint32_t raw[5];
// } tcp_header_t;

// TCP socket_connection state
typedef enum
{
    TCP_STATE_NONE,
    TCP_STATE_LISTEN,
    TCP_STATE_SYN_RCVD,
    TCP_STATE_ESTABLISHED,
    TCP_STATE_SEND_FIRST,
    TCP_STATE_SEND_NEXT,
    TCP_STATE_CLOSING
} tcp_state_t;

#define TCP_MAX_QUEUE_SIZE 40	// size of a retransmission queue

typedef struct
{
	tcp_header_t hdr;
	uint8_t* data;
	uint16_t payload_length;
	uint16_t seq_num;
	uint64_t timer;
	// uint8_t free;
} tcp_frame_send_t;

typedef struct
{
	tcp_frame_send_t frame_send[TCP_MAX_QUEUE_SIZE];
	int front, rear;
	uint8_t size;
} tcp_retransmission_queue_t;

typedef struct
{
    union 
    {
        struct 
        {
            uint32_t nanoseconds;
            uint64_t seconds;
        };
            
        uint32_t raw[3];
    };
} time_stamp_t;

typedef struct 
{
    union
    {
        struct 
        {
            int32_t nanoseconds;
            int64_t seconds;
        };
        uint32_t raw[3];
    }; 
} time_interval_t;


#endif
