#ifndef _UTILS_H_
#define _UTILS_H_

#include <stdio.h>
#include <stdlib.h>

#include "net_types.h"
#include "generic_types.h"

// utils_log workaround to enable logging for debugging
void utils_log(const char *format, ...);
/*
#if defined ENABLE_DEBUG_LOG && !defined WCET_ANALYSIS 
#define utils_log printf
#endif
*/
// functions to calculate the leading zeros of an 8,16 and 32 bit values
uint32_t utils_nlz_32(uint32_t x);
uint16_t utils_nlz_16(uint16_t x);
uint8_t utils_nlz_8(uint8_t x);

// function to compare two mac addresses
uint8_t utils_compare_mac(const mac_address_t* mac1, const  mac_address_t *mac2);

// function to check whether MAC address is a broadcast address
uint8_t utils_compare_mac_broadcast(const  mac_address_t* mac);

// function to check whether two ips are equal
uint8_t utils_compare_ip(const ip_address_t *ip1, const ip_address_t *ip2);

// function to check whether two ips belong to the same network
uint8_t utils_compare_ip_netmask(const ip_address_t *ip1, const ip_address_t *ip2, uint16_t netmask );

// function to check whether an ip is a broacast ip of a network
uint8_t utils_compare_ip_broadcast(const ip_address_t *ip, uint16_t netmask );

// debug functions for printing ip and mac addresses
void utils_print_mac(const mac_address_t* mac);
void utils_print_ip(const ip_address_t* ip);

// utility functions for converting between network an host order
uint16_t utils_htons(uint16_t data);
uint16_t utils_ntohs(uint16_t data);
uint32_t utils_htonl(uint32_t data);
uint32_t utils_ntohl(uint32_t data);


// calculate the sum 2 time intervals and  store in interval
void utils_time_stamp_add(time_stamp_t *ts1, time_stamp_t *ts2, time_interval_t *interval);
// calculate the difference between 2 time intervals and  store in interval
void utils_time_stamp_sub(time_stamp_t *ts1, time_stamp_t *ts2, time_interval_t *interval);
// add a time interval to a timestamp
void utils_time_stamp_add_interval(time_stamp_t *ts, time_interval_t *interval);
// add two time intervals
void utils_time_interval_add(time_interval_t *interval, time_interval_t *add);
// subtract two time intervals
void utils_time_interval_sub(time_interval_t *interval, time_interval_t *sub);
// use this function to correct eventually occured under and overflows in time intervals
void utils_time_interval_correct_of_uf(time_interval_t *interval);
// use this function to correct eventually occured overflows in time stamps
void utils_time_stamp_correct_of(time_stamp_t *ts);


#endif