
#ifndef BUFFER_MANAGEMENT_MMIO_TLSF_H_
#define BUFFER_MANAGEMENT_MMIO_TLSF_H_

#include "../../inc/generic_types.h"
#include "../../inc/buffer_management.h"

#ifdef ARCH_PATMOS
#include <machine/patmos.h>

typedef struct 
{
	uint16_t reserved	:	2;
	uint16_t size_prev  :  14;
	uint16_t alloc  	:	1;
	uint16_t last_block	:	1;
	uint16_t size		:  14;
} mmio_tlsf_block_header_t;

//typedef volatile _IODEV mmio_tlsf_block_header_t * block_header_tlsf_ptr_t;

typedef struct mmio_tlsf_bh_free_t
{
	mmio_tlsf_block_header_t    bh;
	struct mmio_tlsf_bh_free_t volatile _IODEV	*prev;
	struct mmio_tlsf_bh_free_t volatile _IODEV	*next;
} mmio_tlsf_block_header_free_t;

typedef volatile _IODEV mmio_tlsf_block_header_free_t * mmio_tlsf_block_header_free_ptr_t;

typedef struct 
{
	unsigned char fl_lowest_cls;
	unsigned char fl_cls_cnt;
	unsigned char sl_cls_cnt;
    unsigned short fl_free;
	mmio_ptr_t start_address;
    mmio_ptr_t end_address;
    mmio_ptr_16_t sl_free;
    mmio_tlsf_block_header_free_ptr_t *free_list_first;
    mmio_tlsf_block_header_free_ptr_t *free_list_last;
} mmio_tlsf_runtime_config_t;

typedef volatile _IODEV mmio_tlsf_runtime_config_t* mmio_tlsf_runtime_config_prt_t;

#else

typedef struct 
{
	uint16_t reserved	:	2;
	uint16_t size_prev  :  14;
	uint16_t alloc  	:	1;
	uint16_t last_block	:	1;
	uint16_t size		:  14;
} mmio_tlsf_block_header_t;

typedef struct mmio_tlsf_bh_free_t
{
	mmio_tlsf_block_header_t    bh;
	struct mmio_tlsf_bh_free_t volatile	*prev;
	struct mmio_tlsf_bh_free_t volatile	*next;
} mmio_tlsf_block_header_free_t;

typedef volatile mmio_tlsf_block_header_free_t* mmio_tlsf_block_header_free_ptr_t;

typedef struct 
{
	unsigned char fl_lowest_cls;
	unsigned char fl_cls_cnt;
	unsigned char sl_cls_cnt;
    unsigned short fl_free;
	mmio_ptr_t start_address;
    mmio_ptr_t end_address;
    mmio_ptr_16_t sl_free;
    mmio_tlsf_block_header_free_ptr_t *free_list_first;
    mmio_tlsf_block_header_free_ptr_t *free_list_last;
} mmio_tlsf_runtime_config_t;

typedef volatile mmio_tlsf_runtime_config_t* mmio_tlsf_runtime_config_prt_t;

#endif

// initialise TLSF on a given memory space
void * mmio_tlsf_init(buffer_management_config_t *config);

// allocate a buffer 
void * mmio_tlsf_malloc(uint16_t size, void * config_ptr);
// free a buffer again
void mmio_tlsf_free(void * address, void * config_ptr);

// reallocate a part of a buffer, returns NULL if not possible 
void * mmio_tlsf_realloc(void * address, uint16_t size, void * config_ptr);

// debug function to print out the memory content
void mmio_tlsf_memory_dump(void * config_ptr);

#endif
