/*
 * Buffer Management for libnetwork
 * 
 * Authors: Christoph Lehr
 */


#ifndef BUFFER_MANAGEMENT_MMIO_UNIFORM_BUFFER_H_
#define BUFFER_MANAGEMENT_MMIO_UNIFORM_BUFFER_H_

#include "../../inc/net_types.h"
#include "../../inc/buffer_management.h"

#ifdef ARCH_PATMOS
#include <machine/patmos.h>

typedef struct mmio_uniform_bh
{
	struct mmio_uniform_bh volatile _IODEV *next;
} mmio_uniform_block_header_t;

typedef volatile _IODEV mmio_uniform_block_header_t* mmio_uniform_block_header_ptr_t;

typedef struct 
{
	unsigned short buffer_size;
    mmio_ptr_t mgmt_address;
	mmio_ptr_t start_address;
    mmio_ptr_t end_address;
    mmio_uniform_block_header_ptr_t free_list_first;
    mmio_uniform_block_header_ptr_t free_list_last;
} mmio_uniformm_runtime_config_t;

typedef mmio_uniformm_runtime_config_t volatile _IODEV * mmio_uniformm_runtime_config_ptr_t;

#else

typedef struct mmio_uniform_bh
{
	struct mmio_uniform_bh volatile *next;
} mmio_uniform_block_header_t;

typedef volatile mmio_uniform_block_header_t* mmio_uniform_block_header_ptr_t;

typedef struct 
{
	unsigned short buffer_size;
    mmio_ptr_t mgmt_address;
	mmio_ptr_t start_address;
    mmio_ptr_t end_address;
    mmio_uniform_block_header_ptr_t free_list_first;
    mmio_uniform_block_header_ptr_t free_list_last;
} mmio_uniformm_runtime_config_t;

typedef mmio_uniformm_runtime_config_t volatile * mmio_uniformm_runtime_config_ptr_t;
#endif

// initialise unfirom buffer management
void * mmio_uniform_buffer_init(buffer_management_config_t *config);

// allocate a buffer
void * mmio_uniform_buffer_malloc(void * config_ptr);
// free a buffer again
void mmio_uniform_buffer_free(void * buffer, void * config_ptr);

// reallocate part of a buffer 
void * mmio_uniform_buffer_realloc(void * buffer, uint16_t size, void * config_ptr);

#endif
