
#ifndef BUFFER_MANAGEMENT_TLSF_H__
#define BUFFER_MANAGEMENT_TLSF_H__

#include "../../inc/generic_types.h"
#include "../../inc/buffer_management.h"

typedef struct 
{
	uint16_t reserved	:	2;
	uint16_t size_prev  :  14;
	uint16_t alloc  	:	1;
	uint16_t last_block	:	1;
	uint16_t size		:  14;
} block_header_tlsf_t;

//typedef volatile block_header_tlsf_t * block_header_tlsf_ptr_t;

typedef struct bh_free_t
{
	block_header_tlsf_t    bh;
	struct bh_free_t volatile *prev;
	struct bh_free_t volatile *next;
} block_header_free_tlsf_t;

typedef volatile block_header_free_tlsf_t * block_header_free_tlsf_ptr_t;

typedef struct 
{
	unsigned char fl_lowest_cls;
	unsigned char fl_cls_cnt;
	unsigned char sl_cls_cnt;
    unsigned short fl_free;
	ptr_t start_address;
    ptr_t end_address;
    ptr_16_t sl_free;
    block_header_free_tlsf_ptr_t *free_list_first;
    block_header_free_tlsf_ptr_t *free_list_last;
} tlsf_runtime_config_t;

typedef volatile tlsf_runtime_config_t* tlsf_runtime_config_prt_t;

// initialise TLSF on a given memory space
void * tlsf_init(buffer_management_config_t *config);

// allocate a buffer 
void * tlsf_malloc(uint16_t size, void * config_ptr);
// free a buffer again
void tlsf_free(void * address, void * config_ptr);

// reallocate a part of a buffer, returns NULL if not possible 
void * tlsf_realloc(void * address, uint16_t size, void * config_ptr);

// debug function to print out the memory content
void tlsf_memory_dump(void * config_ptr);

#endif
