/*
 * Buffer Management for libnetwork
 * 
 * Authors: Christoph Lehr
 */


#ifndef _INCLUDE_UNIFORM_BUFFER_H_
#define _INCLUDE_UNIFORM_BUFFER_H_

#include "../../inc/net_types.h"
#include "../../inc/buffer_management.h"

#ifdef ARCH_PATMOS
#include <machine/patmos.h>
#endif

typedef struct bh_uniform
{
	struct bh_uniform volatile *next;
} block_header_uniform_t;

typedef volatile block_header_uniform_t* block_header_uniform_ptr_t;

typedef struct 
{
	unsigned short buffer_size;
    ptr_t mgmt_address;
	ptr_t start_address;
    ptr_t end_address;
    block_header_uniform_ptr_t free_list_first;
    block_header_uniform_ptr_t free_list_last;
} uniformm_runtime_config_t;

typedef uniformm_runtime_config_t volatile * uniformm_runtime_config_ptr_t;

// initialise unfirom buffer management
void * uniform_buffer_init(buffer_management_config_t *config);

// allocate a buffer
void * uniform_buffer_malloc(void * config_ptr);
// free a buffer again
void uniform_buffer_free(void * buffer, void * config_ptr);

// reallocate part of a buffer 
void * uniform_buffer_realloc(void * buffer, uint16_t size, void * config_ptr);

#endif
