/*
 * IO functions to read froma and write to the Etherent MAC 
 * 
 * Authors: Luca Pezzarossa (lpez@dtu.dk)
 *          Jakob Kenn Toft
 *          Jesper Lønbæk
 *          Russell Barnes
 *          Christoph Lehr
 */

#include <stdio.h>
#include "../../inc/generic_types.h"

#ifndef MMIO_WCET_MAX_WRITE_DATA_SIZE
    // write to memory in word bursts
    // 379.5 = 1520 Bytes (Max MTU) / 4 (Bytes per Word)
    // The function will be executed twice for each frame
    // one time for the headers, one time for the data 
    // therefore the total cycles get halfed :380/2 = 190
    #define MMIO_WCET_MAX_WRITE_DATA_SIZE 190 
#endif

#ifndef MMIO_WCET_MAX_READ_DATA_SIZE
    // Maximum length of an Ethernet frame is 1518 Bytes
    // 379.5 = 1500 Bytes (Max MTU) / 4 (Bytes per Word)
    #define MMIO_WCET_MAX_READ_DATA_SIZE 380
#endif



void mmio_wr_8(mmio_ptr_t address, uint16_t offset ,uint8_t data);

uint8_t mmio_rd_8(mmio_ptr_t address, uint16_t offset);

uint16_t mmio_rd_16(mmio_ptr_t address, uint16_t offset);

void mmio_wr_16(mmio_ptr_t address, uint16_t offset ,uint8_t data);

void mmio_wr_32(mmio_ptr_t address, uint16_t offset , uint32_t data);

uint32_t mmio_rd_32(mmio_ptr_t address, uint16_t offset);

void mmio_wr_block(mmio_ptr_t address, uint8_t *data, uint16_t len);

void mmio_rd_block(mmio_ptr_t address, uint8_t *data, uint16_t len);