/*
 * ARP section 
 * 
 * Authors: Luca Pezzarossa (lpez@dtu.dk)
 *          Jakob Kenn Toft
 *          Jesper Lønbæk
 *          Russell Barnes
 *          Christoph Lehr
 */

#ifndef _ARP_H_
#define _ARP_H_

#include "../../inc/net_types.h"
#include "../../inc/generic_types.h"

#ifndef ARP_TABLE_SIZE
    #define ARP_TABLE_SIZE 8
#endif

///////////////////////////////////////////////////////////////
//Functions related to the ARP table
///////////////////////////////////////////////////////////////

void arp_init();

//This function insert a new entry in the ARP IP/MAC table. If an entry was already there the fields are updated. If we are out of space, the first entry is replaced.
network_error_t arp_table_new_entry(ip_address_t *ip_addr, mac_address_t *mac_addr);

//This function searches in the ARP table for the given IP. If it exists it returns 1 and the MAC. If not it returns 0.
network_error_t arp_table_search(ip_address_t *ip_addr, mac_address_t *mac_addr);

//This ugly function prints the ARP table for debug purposes.
void arp_table_print();


///////////////////////////////////////////////////////////////
//Support functions related to the ARP protocol
///////////////////////////////////////////////////////////////

//This function process a received ARP package. If it is a request and we are the destination (IP) it reply the ARP request and returns 1. If it is a reply and we are the destination (IP) it inserts an entry in the ARP table and returns 2. Otherwise it returns 0.
network_error_t arp_handle_message(uint8_t *buffer, uint16_t size, ip_address_t* host_ip, mac_address_t* host_mac, arp_msg_t *arp_response);

//This function builds an ARP request packet starting at tx_addr and targeting the target_ip
network_error_t arp_build_request(ip_address_t *target_ip, ip_address_t* host_ip, mac_address_t* host_mac, arp_msg_t * request);

#endif
