/*
 * Ethernet Protocol section
 * 
 * Authors: Luca Pezzarossa (lpez@dtu.dk)
 *          Jakob Kenn Toft
 *          Jesper Lønbæk
 *          Russell Barnes
 * 			Lehr Christoph
 */

#include <stdio.h>
#include "ethernet.h"
#include "../../inc/net_types.h"
#include "../../inc/generic_types.h"


//Write Mac Header to Buffer, returns length of written header length
network_error_t ethernet_build_frame(mac_address_t *dst_mac, mac_address_t *src_mac, ethertype_t protocol, uint8_t * buffer)
{
	ethernet_header_t *eth_hdr = (ethernet_header_t * )buffer;

	// Copy MAC Addresses
	#pragma loopbound min 6 max 6
	for(unsigned char i=0; i<6; i++)
	{
		eth_hdr->dest_mac[i] = dst_mac->address[i];
		eth_hdr->src_mac[i] = src_mac->address[i];
	}

	eth_hdr->eth_type=protocol;

	return NET_ERR_OK;
};

network_error_t ethernet_build_vlan_frame(mac_address_t *dst_mac, mac_address_t *src_mac, ethertype_t protocol, vlan_header_t *vlan_cfg, uint8_t * buffer)
{
	ethernet_header_vlan_t *eth_hdr = (ethernet_header_vlan_t * )buffer;

	// Copy MAC Addresses
	#pragma loopbound min 6 max 6
	for(unsigned char i=0; i<6; i++)
	{
		eth_hdr->dest_mac[i] = dst_mac->address[i];
		eth_hdr->src_mac[i]  = src_mac->address[i];
	}
	eth_hdr->vlan_eth_type = ETHERTYPE_VLAN;
	eth_hdr->pcp = vlan_cfg->pcp;
	eth_hdr->dei = vlan_cfg->dei;
	eth_hdr->vid = vlan_cfg->vid;
	eth_hdr->eth_type=protocol;

	return NET_ERR_OK;
};

network_error_t ethernet_handle_frame(uint8_t *buffer, uint16_t size, ethertype_t *protocol, mac_address_t* src_mac, mac_address_t* dst_mac)
{
	// convert buffer to ethernet frame header
	ethernet_header_t *eth_hdr = (ethernet_header_t *) buffer; 

	if(size < ETHERNET_MIN_LENGTH)
	{
		return NET_ERR_ETH_FRAME_LENGTH;
	}

	*protocol = eth_hdr->eth_type;

	src_mac->length =6;
	dst_mac->length =6;
	#pragma loopbound min 6 max 6 
	for(uint8_t i=0; i < 6; i++)
	{
		src_mac->address[i] = eth_hdr->src_mac[i];
		dst_mac->address[i] = eth_hdr->dest_mac[i];
	}

	return NET_ERR_OK;
}

network_error_t ethernet_handle_vlan_frame(uint8_t *buffer, uint16_t size, ethertype_t *protocol, mac_address_t* src_mac, mac_address_t* dst_mac, network_layer2_config_t *l2_cfg)
{
	// convert buffer to ethernet frame header
	ethernet_header_vlan_t *eth_hdr = (ethernet_header_vlan_t *) buffer; 

	if(eth_hdr->vlan_eth_type != ETHERTYPE_VLAN)
	{
		return NET_ERR_ETH_NO_VLAN;
	}

	if(l2_cfg != NULL && l2_cfg->l2_type == NETWORK_L2_ETHERNET_VLAN && eth_hdr->vid != l2_cfg->vlan.vid)
	{
		return NET_ERR_ETH_WRONG_VLAN;
	}

	*protocol = eth_hdr->eth_type;

	return NET_ERR_OK;
}
