/*
 * Ethernet Protocol section
 * 
 * Authors: Luca Pezzarossa (lpez@dtu.dk)
 *          Jakob Kenn Toft
 *          Jesper Lønbæk
 *          Russell Barnes
 */

#ifndef _MAC_H_
#define _MAC_H_

#include "../../inc/net_types.h"
#include "../../inc/generic_types.h"
#include "../../inc/network.h"

#define ETHERNET_HEADER_LEN 14
#define ETHERNET_HEADER_VLAN_LEN 18

#define ETHERNET_MIN_LENGTH 64

void ethernet_init(void);

const unsigned char* ethernet_get_my_mac();

//This is used configure vlan of the host 
void ethernet_set_vlan(unsigned char enable, unsigned short vlan_id);

//Write Mac Header to Buffer
network_error_t ethernet_build_frame(mac_address_t *dst_mac, mac_address_t *src_mac, ethertype_t protocol, uint8_t * buffer);
network_error_t ethernet_build_vlan_frame(mac_address_t *dst_mac, mac_address_t *src_mac, ethertype_t protocol, vlan_header_t *vlan_cfg, uint8_t * buffer);

// Parse Ethernet frame from buffer
network_error_t ethernet_handle_frame(uint8_t *buffer, uint16_t size, ethertype_t *protocol, mac_address_t* src_mac, mac_address_t* dst_mac);
network_error_t ethernet_handle_vlan_frame(uint8_t *buffer, uint16_t size, ethertype_t *protocol, mac_address_t* src_mac, mac_address_t* dst_mac, network_layer2_config_t *l2_cfg);

#endif
