/*
 * IPv4 section 
 * 
 * Authors: Luca Pezzarossa (lpez@dtu.dk)
 *          Jakob Kenn Toft
 *          Jesper Lønbæk
 *          Russell Barnes
 *          Christoph Lehr
 */

#ifndef _IPV4_H_
#define _IPV4_H_

#include "../../inc/net_types.h"
#include "../../inc/generic_types.h"

#define IPv4_IDENTIFICATION_SEED 0x1234
#define IPv4_IDENTIFICATION_INC  0x4321

// An ethernet frame has a minimum size of 64 bytes
//Ethernet Header = 14 Bytes => min size = 50 bytes
#define IP_MIN_PACKET_SIZE 50

//Write Mac Header to Buffer, returns length of written header length. Header offset is the offset to tx address
network_error_t ipv4_build_datagram(ip_address_t *dst_ip, ip_address_t *src_ip, ip_protocol_t protocol, uint8_t *buffer, uint16_t len);

network_error_t ipv4_handle_packet(uint8_t*buffer, uint16_t size, ip_protocol_t *protocol, ip_address_t *src_ip, ip_address_t *dst_ip, uint16_t *payload_length);
///////////////////////////////////////////////////////////////
//Support functions related to the IPv4 protocol
///////////////////////////////////////////////////////////////

//This function compute and returns the IP header checksum. The function ignore the field checksum. Uses the Buffer Address as input
unsigned short ipv4_compute_checksum(ipv4_header_t *ipv4_header);

//This function verify the IP header checksum. If the checksum is correct it returns 1, otherwise it returns 0.
char ipv4_verify_checksum(ipv4_header_t *ipv4_header);

#endif
