
/*
 * UDP section
 * 
 * Authors: Christoph Lehr
 */

#ifndef _UDP_H_
#define _UDP_H_

#include "../../inc/net_types.h"
#include "../../inc/generic_types.h"

#ifndef UDP_WCET_MAX_DATA_SIZE
   // Implementation does not support Ethernet jumbo frames 
	// or IPv4 Fragmentation. Max Ethernet user data 1500 Byte
	// minus IPv4 Header 20 Bytes and UDP Header 8 Bytes, therefore
	// the maximum payload of UDP is 1472 Byte, which comes down to
	// 368 32 bit words. (1500 -20-8) /4 = 368
   #define UDP_WCET_MAX_DATA_SIZE 368 
#endif

// An ethernet frame has a minimum size of 64 bytes
// Ethernet Header + IPv4 = 30 Bytes => min size = 34 bytes
#define UDP_MIN_PACKET_SIZE 34

// compute the udp checksum for a packet
unsigned short udp_compute_checksum(udp_header_t *udp_hdr, ip_address_t *dest_ip, ip_address_t *src_ip, unsigned char *data, unsigned short data_length);

//This function computes the UDP checksum and returns whether it is correct
unsigned char udp_verify_checksum(udp_header_t *udp_hdr, ip_address_t *dest_ip, ip_address_t *src_ip, unsigned char *data, unsigned short data_length);

//This function sends an UDP packet to the dstination IP.
network_error_t udp_build_datagram(
	ip_address_t *destination_ip, uint16_t destination_port, 
	ip_address_t *source_ip, uint16_t source_port, 
	uint8_t *data, uint16_t data_length, uint8_t *buffer);

// This function parses a received udp packet and checks wether the checksum is correct
network_error_t udp_handle_datagram(uint8_t *buffer, uint16_t size, 
	ip_address_t *destination_ip, uint16_t *destination_port, 
	ip_address_t *source_ip, uint16_t *source_port);

#endif
