#ifndef _LIBNETWORK_SERVICES_TIME_SYNCHRONIATION_PTP_H_
#define _LIBNETWORK_SERVICES_TIME_SYNCHRONIATION_PTP_H_


#include "../../../inc/generic_types.h"

typedef enum
{
    PTP_MSG_TYPE_SYNC                   = 0x0,
    PTP_MSG_TYPE_DELAY_REQ              = 0x1,
    PTP_MSG_TYPE_PDELAY_REQ             = 0x2,
    PTP_MSG_TYPE_PDELAY_RESP            = 0x3,
//  0x4-0x7 are reserved        
    PTP_MSG_TYPE_FOLLOW_UP              = 0x8,
    PTP_MSG_TYPE_DELAY_RESP             = 0x9,
    PTP_MSG_TYPE_PDELAY_RESP_FOLLOW_UP  = 0xA,
    PTP_MSG_TYPE_ANNOUNCE               = 0xB,
    PTP_MSG_TYPE_SIGNALLING             = 0xC,
    PTP_MSG_TYPE_MANAGEMENT             = 0xD,
//  0xE-0xF are reserved
} ptp_msg_type_t;

typedef struct __attribute__((__packed__)) 
{
    uint8_t id[8];
}ptp_clock_identity_t;

typedef struct __attribute__((__packed__)) 
{
    ptp_clock_identity_t clockIdentity;
    uint16_t portNumber;
}ptp_port_identity_t; 

typedef union 
{
    struct {
                                            // byte 0
        
        const uint16_t reserved_2:1;        // Bit 7
        uint16_t profile_specific_2:1;      // Bit 6
        uint16_t profile_specific_1:1;      // Bit 5
        const uint16_t reserved_1:2;        // Bit 3 4 
        uint16_t unicastFlag:1;             // Bit 2
        uint16_t twoStepFlag:1;             // Bit 1 
        uint16_t alternateMasterFlag:1;     // Bit 0
        
                                            // byte 1
        const uint16_t reserved_3:1;        // Bit 7
        uint16_t synchronizationUncertain:1;// Bit 6
        uint16_t frequencyTraceable:1;      // Bit 5
        uint16_t timeTraceable:1;           // Bit 4
        uint16_t ptpTimescale:1;            // Bit 3
        uint16_t currentUtcOffsetValid:1;   // Bit 2
        uint16_t leap59:1;                  // Bit 1
        uint16_t leap61:1;                  // Bit 0
    };
    uint16_t raw;
} ptp_flags_t;

typedef enum
{
    PTP_CLOCK_ACCURACY_1_PS         =   0x17,
    PTP_CLOCK_ACCURACY_2_5_PS       =   0x18,
    PTP_CLOCK_ACCURACY_10_PS        =   0x19,
    PTP_CLOCK_ACCURACY_25_PS        =   0x1A,
    PTP_CLOCK_ACCURACY_100_PS       =   0x1B,
    PTP_CLOCK_ACCURACY_250_PS       =   0x1C,
    PTP_CLOCK_ACCURACY_1_NS         =   0x1D,
    PTP_CLOCK_ACCURACY_2_5_NS       =   0x1E,
    PTP_CLOCK_ACCURACY_10_NS        =   0x1F,
    PTP_CLOCK_ACCURACY_25_NS        =   0x20,
    PTP_CLOCK_ACCURACY_100_NS       =   0x21,
    PTP_CLOCK_ACCURACY_250_NS       =   0x22,
    PTP_CLOCK_ACCURACY_1_US         =   0x23,
    PTP_CLOCK_ACCURACY_2_5_US       =   0x24,
    PTP_CLOCK_ACCURACY_10_US        =   0x25,
    PTP_CLOCK_ACCURACY_25_US        =   0x26,
    PTP_CLOCK_ACCURACY_100_US       =   0x27,
    PTP_CLOCK_ACCURACY_250_US       =   0x28,
    PTP_CLOCK_ACCURACY_1_MS         =   0x29,
    PTP_CLOCK_ACCURACY_2_5_MS       =   0x2A,
    PTP_CLOCK_ACCURACY_10_MS        =   0x2B,
    PTP_CLOCK_ACCURACY_25_MS        =   0x2C,
    PTP_CLOCK_ACCURACY_100_MS       =   0x2D,
    PTP_CLOCK_ACCURACY_250_MS       =   0x2E,
    PTP_CLOCK_ACCURACY_1_S          =   0x2F,
    PTP_CLOCK_ACCURACY_10_S         =   0x30,
    PTP_CLOCK_ACCURACY_GT_10_S      =   0x31,
} ptp_clock_accuracy_t;

typedef struct __attribute__((__packed__))  
{
    uint8_t clockClass;
    ptp_clock_accuracy_t clockAccuracy;
    uint16_t offsetScaledLogVariance;
} ptp_clock_quality_t;

typedef enum
{
    PTP_TIME_SOURCE_NONE                =   0x00,
    PTP_TIME_SOURCE_ATOMIC_CLOCK        =   0x10,
    PTP_TIME_SOURCE_GNSS                =   0x20,
    PTP_TIME_SOURCE_TERRESTRIAL_RADIO   =   0x30,
    PTP_TIME_SOURCE_SERIAL_TIME_CODE    =   0x39,
    PTP_TIME_SOURCE_PTP                 =   0x40,
    PTP_TIME_SOURCE_NTP                 =   0x50,
    PTP_TIME_SOURCE_HAND_SET            =   0x60,
    PTP_TIME_SOURCE_OTHER               =   0x90,
    PTP_TIME_SOURCE_INTERNAL_OSCILLATOR =   0xA0,
} ptp_time_source_t;


// attribute required because clang adds dunno why 2 byte padding
typedef struct __attribute__((__packed__))  {
    uint8_t majorSdoId           :4;
    uint8_t msg_type             :4;
    uint8_t minorVersionPTP      :4;
    uint8_t majorVersionPTP      :4;
    uint16_t messageLength;
    uint8_t domainNumber;
    uint8_t minorSdoId;
    ptp_flags_t flagField;
    int64_t correctionField;
    uint32_t messageTypeSpecific;
    ptp_port_identity_t sourcePortIdentity;
    uint16_t sequenceId;
    uint8_t controlField;
    uint8_t logMessageInterval;

}ptp_header_t;

typedef struct __attribute__((__packed__))
{
    uint64_t seconds:48;
    uint32_t nanoseconds;
} ptp_timestamp_t;

typedef struct __attribute__((__packed__)) 
{
    ptp_header_t header;
    ptp_timestamp_t originTimestamp;
    uint16_t currentUtcOffset;
    const uint8_t reserved;
    uint8_t grandmasterPriority1;
    ptp_clock_quality_t grandmasterClockQuality;
    uint8_t grandmasterPriority2;
    ptp_clock_identity_t grandmasterIdentity;
    uint16_t stepsRemoved;
    ptp_time_source_t timeSource;
} ptp_announce_msg_t;

typedef struct __attribute__((__packed__))
{
    ptp_header_t header;
    ptp_timestamp_t originTimestamp;
} ptp_sync_msg_t;

typedef struct __attribute__((__packed__))
{
    ptp_header_t header;
    ptp_timestamp_t originTimestamp;
} ptp_delay_req_msg_t;


typedef struct __attribute__((__packed__))
{
    ptp_header_t header;
    ptp_timestamp_t preciseOriginTimestamp;
    struct
    {
        uint16_t tlvType;
        uint16_t lengthField;
        uint8_t organizationId[3];
        uint8_t organizationSubType[3];
        uint8_t cumulativeScaledRateOffset[4];
        uint16_t gmTimeBaseIndicator;
        uint8_t lastGmPhaseChange[12];
        uint64_t scaledLastGmFreqChange;
    } tlv;
} ptp_follow_up_msg_t;

typedef struct __attribute__((__packed__))
{
    ptp_header_t header;
    ptp_timestamp_t preciseOriginTimestamp;
    ptp_port_identity_t requestingPortIdentity;
} ptp_delay_resp_msg_t;

typedef struct __attribute__((__packed__))
{
    ptp_header_t header;
    ptp_timestamp_t originTimestamp;
    const uint8_t reserved[10];
} ptp_pdelay_req_msg_t;

typedef struct __attribute__((__packed__))
{
    ptp_header_t header;
    ptp_timestamp_t requestReceiptTimestamp;
    ptp_port_identity_t requestingPortIdentity;
} ptp_pdelay_resp_msg_t;

typedef struct __attribute__((__packed__))
{
    ptp_header_t header;
    ptp_timestamp_t responseOriginTimestamp;
    ptp_port_identity_t requestingPortIdentity;
} ptp_pdelay_resp_follow_up_msg_t;

typedef union 
{
    ptp_header_t            header;
    ptp_announce_msg_t      announce;
    ptp_sync_msg_t          sync;
    ptp_delay_req_msg_t     delay_req;
    ptp_follow_up_msg_t     follow_up;
    ptp_delay_resp_msg_t    delay_resp;
    ptp_pdelay_req_msg_t    pdelay_req;
    ptp_pdelay_resp_msg_t   pdelay_resp;
    ptp_pdelay_resp_follow_up_msg_t pdelay_resp_follow_up;
} ptp_msg_t;

typedef enum
{
    PTP_PORT_STATE_MASTER,
    PTP_PORT_STATE_SLAVE,
    PTP_PORT_STATE_PASSIVE,
    PTP_PORT_STATE_DISABLED,
} ptp_port_state_t;

#endif
