#include <stdio.h>
#include <stdlib.h>

#include "../inc/utils.h"
#include <machine/exceptions.h>
#include <machine/patmos.h>


#define SECONDS_TO_NANOSECONDS 1000000000

   // dummy logging function as workaround to have WCET analyseabliltiy and logging for debugging
void utils_log(const char *format, ...)
{   
#if defined ENABLE_DEBUG_LOG && !defined WCET_ANALYSIS 
    /*
    unsigned intr_en = EXC_STATUS & 0x1;
    if(intr_en)
    {
        CLI();
    }
    */
    /* Declare a va_list type variable */
    va_list myargs;

    /* Initialise the va_list variable with the ... after fmt */

    va_start(myargs, format);

    /* Forward the '...' to vprintf */
    vprintf(format, myargs);

    /* Clean up the va_list */
    va_end(myargs);
    /*
    if(intr_en)
    {
        SEI();
    }*/
#endif
    return;
}


uint32_t utils_nlz_32(uint32_t x)
{
    // Algorithm from: https://embeddedgurus.com/state-space/2014/09/fast-deterministic-and-portable-counting-leading-zeros/
    static uint8_t const clz_lkup[] = {
        32U, 31U, 30U, 30U, 29U, 29U, 29U, 29U,
        28U, 28U, 28U, 28U, 28U, 28U, 28U, 28U,
        27U, 27U, 27U, 27U, 27U, 27U, 27U, 27U,
        27U, 27U, 27U, 27U, 27U, 27U, 27U, 27U,
        26U, 26U, 26U, 26U, 26U, 26U, 26U, 26U,
        26U, 26U, 26U, 26U, 26U, 26U, 26U, 26U,
        26U, 26U, 26U, 26U, 26U, 26U, 26U, 26U,
        26U, 26U, 26U, 26U, 26U, 26U, 26U, 26U,
        25U, 25U, 25U, 25U, 25U, 25U, 25U, 25U,
        25U, 25U, 25U, 25U, 25U, 25U, 25U, 25U,
        25U, 25U, 25U, 25U, 25U, 25U, 25U, 25U,
        25U, 25U, 25U, 25U, 25U, 25U, 25U, 25U,
        25U, 25U, 25U, 25U, 25U, 25U, 25U, 25U,
        25U, 25U, 25U, 25U, 25U, 25U, 25U, 25U,
        25U, 25U, 25U, 25U, 25U, 25U, 25U, 25U,
        25U, 25U, 25U, 25U, 25U, 25U, 25U, 25U,
        24U, 24U, 24U, 24U, 24U, 24U, 24U, 24U,
        24U, 24U, 24U, 24U, 24U, 24U, 24U, 24U,
        24U, 24U, 24U, 24U, 24U, 24U, 24U, 24U,
        24U, 24U, 24U, 24U, 24U, 24U, 24U, 24U,
        24U, 24U, 24U, 24U, 24U, 24U, 24U, 24U,
        24U, 24U, 24U, 24U, 24U, 24U, 24U, 24U,
        24U, 24U, 24U, 24U, 24U, 24U, 24U, 24U,
        24U, 24U, 24U, 24U, 24U, 24U, 24U, 24U,
        24U, 24U, 24U, 24U, 24U, 24U, 24U, 24U,
        24U, 24U, 24U, 24U, 24U, 24U, 24U, 24U,
        24U, 24U, 24U, 24U, 24U, 24U, 24U, 24U,
        24U, 24U, 24U, 24U, 24U, 24U, 24U, 24U,
        24U, 24U, 24U, 24U, 24U, 24U, 24U, 24U,
        24U, 24U, 24U, 24U, 24U, 24U, 24U, 24U,
        24U, 24U, 24U, 24U, 24U, 24U, 24U, 24U,
        24U, 24U, 24U, 24U, 24U, 24U, 24U, 24U
    };
    uint32_t n;
    if (x >= (1U << 16)) {
        if (x >= (1U << 24)) {
            n = 24U;
        }
        else {
            n = 16U;
        }
    }
    else {
        if (x >= (1U << 8)) {
            n = 8U;
        }
        else {
            n = 0U;
        }
    }
    return (uint32_t)clz_lkup[x >> n] - n;

    /*
	uint16_t n = 32; 
	if ((x>>16) != 0) { n=n-16; x = x>>16; } 
	if ((x>>8 ) != 0) { n=n- 8; x = x>> 8; } 
	if ((x>>4 ) != 0) { n=n- 4; x = x>> 4; } 
	if ((x>>2 ) != 0) { n=n- 2; x = x>> 2; } 
	if ((x>>1 ) != 0) { return n-2; } 
    return n - x; 	
    */
}

uint16_t utils_nlz_16(uint16_t x)
{ 
    // reduce leading zeros by 16 to convert from 32 to 16 bit
    return (uint16_t) utils_nlz_32(x)-16 ;
    /*
	uint16_t n = 16; 
	if ((x>>8 ) != 0) { n=n- 8; x = x>> 8; } 
	if ((x>>4 ) != 0) { n=n- 4; x = x>> 4; } 
	if ((x>>2 ) != 0) { n=n- 2; x = x>> 2; } 
	if ((x>>1 ) != 0) { return n-2; } 
    return n - x; 	
    */
}

uint8_t utils_nlz_8(uint8_t x)
{ 
    /*
	uint16_t n = 8; 
	if ((x>>4 ) != 0) { n=n- 4; x = x>> 4; } 
	if ((x>>2 ) != 0) { n=n- 2; x = x>> 2; } 
	if ((x>>1 ) != 0) { return n-2; } 
	return n - x; 	
    */
    // reduce leading zeros by 16 to convert from 32 to 8 bit
    return (uint8_t) utils_nlz_32(x)-24 ;
}

uint8_t utils_compare_mac(const  mac_address_t* mac1, const  mac_address_t *mac2)
{
    if(mac1->length != mac2->length)
    {
        return ~0;
    }
    
    uint8_t equal = 1;
    #pragma loopbound min 1 max 8
    for(uint8_t i=0; i< mac1->length; i++)
    {
        equal &= mac1->address[i] == mac2->address[i];
        if(equal == 0) return ~0;
    }

    return 0;
}

uint8_t utils_compare_mac_broadcast(const  mac_address_t* mac)
{
    uint8_t equal = 1;

    #pragma loopbound min 1 max 8
    for(uint8_t i=0; i< mac->length; i++)
    {
        equal &= mac->address[i] == 0xFF;
        if(equal == 0) return ~0;
    }

    return 0;
}

uint8_t utils_compare_ip(const ip_address_t *ip1,const  ip_address_t *ip2)
{
    if(ip1->version != ip2->version)
    {
        return ~0;
    }

    if(ip1->version == IP_VERSION_4)
    {
        uint8_t equal = 1;
        #pragma loopbound min 1 max 4
        for (unsigned short i = 0; i < 4; i++)
        {
            equal &= ip1->ipv4_addr[i] == ip2->ipv4_addr[i];
            if(equal == 0) return ~0;
        }
    }
    else
    {
        uint8_t equal = 1;
        #pragma loopbound min 1 max 8
        for (unsigned short i = 0; i < 8; i++)
        {
            equal &= ip1->ipv6_addr[i] == ip2->ipv6_addr[i];
            if(equal == 0) return ~0;
        }
    }

    return 0;
}

uint8_t utils_compare_ip_broadcast(const ip_address_t *ip, const uint16_t netmask )
{

    if(ip->version == IP_VERSION_4)
    {
        ip_address_t broadcast_ip;
        broadcast_ip.version = IP_VERSION_4;
        uint32_t subnetmask = (1 << netmask)-1;
        #pragma loopbound min 4 max 4
        for (uint8_t i = 0; i < 4; i++)
        {
            uint8_t mask = (uint8_t) (subnetmask >> 8*i);
            broadcast_ip.ipv4_addr[i] = ip->ipv4_addr[i] & mask;
        }

        uint32_t broadcast_mask = (1 << (32-netmask))-1;
        broadcast_mask <<= netmask;
        #pragma loopbound min 4 max 4
        for (uint8_t i = 0; i < 4; i++)
        {
            uint8_t mask = (uint8_t) (broadcast_mask >> 8*i);
            broadcast_ip.ipv4_addr[i] |= mask;
        }

        return utils_compare_ip(ip, &broadcast_ip);
    }
    else
    {
        // IPv6 does not have broadcast addresses
        return ~0;
    }

    return 0;
}

uint8_t utils_compare_ip_netmask(const ip_address_t *ip1, const ip_address_t *ip2, uint16_t netmask )
{
    if(ip1->version != ip2->version)
    {
        return ~0;
    }

    if(ip1->version == IP_VERSION_4)
    {
        uint8_t equal = 1;
        uint32_t subnetmask = (1 << netmask)-1;
        #pragma loopbound min 1 max 4
        for (uint8_t i = 0; i < 4; i++)
        {
            uint8_t mask = (uint8_t) (subnetmask >> 8*i);
            equal &= (ip1->ipv4_addr[i] & mask) == (ip2->ipv4_addr[i] & mask);
            if(equal == 0) return ~0;
        }
    }
    else
    {
        uint8_t equal = 1;
        uint64_t subnetmask = 0xFFFFFFFFFFFFFFFF;
        if(netmask < 63)
        {
            subnetmask = (1 << netmask)-1;
        }    

        #pragma loopbound min 1 max 4
        for (unsigned short i = 0; i < 4; i++)
        {
            uint16_t mask = (uint16_t) (subnetmask >> 16*i);
            equal &= (ip1->ipv6_addr[i] & mask) == (ip2->ipv6_addr[i] & mask);
            if(equal == 0) return ~0;
        }
        
        if(netmask >= 64)
        {
            subnetmask = (1 << (netmask-64))-1;
            #pragma loopbound min 1 max 4
            for (unsigned short i = 0; i < 4; i++)
            {
                uint16_t mask = (uint16_t) (subnetmask >> 16*i);
                equal &= (ip1->ipv6_addr[4+i] & mask) == (ip2->ipv6_addr[4+i] & mask);
                if(equal == 0) return ~0;
            }
        }
    }

    return 0;
}

void utils_print_mac(const mac_address_t* mac){
    if(mac->length == 6)
    {
	    utils_log("%02X:%02X:%02X:%02X:%02X:%02X", 
            mac->address[0], mac->address[1], mac->address[2], 
            mac->address[3], mac->address[4], mac->address[5]);
    }
    else if(mac->length == 8)
    {
        utils_log("%02X:%02X:%02X:%02X:%02X:%02X:%02X:%02X", 
            mac->address[0], mac->address[1], mac->address[2], mac->address[3], 
            mac->address[4], mac->address[5], mac->address[6], mac->address[7]);
    }

	return;
}

void utils_print_ip(const ip_address_t* ip){
    if(ip->version == IP_VERSION_4)
    {
	    utils_log("%u.%u.%u.%u", 
            ip->ipv4_addr[0], ip->ipv4_addr[1], ip->ipv4_addr[2], ip->ipv4_addr[3]);
    }
    else if(ip->version == IP_VERSION_6)
    {
        utils_log("%04X:%04X:%04X:%04X:%04X:%04X:%04X:%04X", 
            ip->ipv6_addr[0], ip->ipv6_addr[1], ip->ipv6_addr[2], ip->ipv6_addr[3], 
            ip->ipv6_addr[4], ip->ipv6_addr[5], ip->ipv6_addr[6], ip->ipv6_addr[7]);
    }

	return;
}


void utils_time_stamp_add(time_stamp_t *ts1, time_stamp_t *ts2, time_interval_t *interval)
{

    interval->nanoseconds   = ts1->nanoseconds;
    interval->seconds       = ts1->seconds;
    interval->nanoseconds  += ts2->nanoseconds;
    interval->seconds      += ts2->seconds;

    utils_time_interval_correct_of_uf(interval);
}

void utils_time_stamp_sub(time_stamp_t *ts1, time_stamp_t *ts2, time_interval_t *interval)
{

    interval->nanoseconds   = ts1->nanoseconds;
    interval->seconds       = ts1->seconds;
    interval->nanoseconds  -= ts2->nanoseconds;
    interval->seconds      -= ts2->seconds;

    if(ts2->seconds > ts1->seconds)
    {
        interval->seconds++;
        interval->nanoseconds = interval->nanoseconds - SECONDS_TO_NANOSECONDS;
    }

    utils_time_interval_correct_of_uf(interval);

}


void utils_time_stamp_add_interval(time_stamp_t *ts, time_interval_t *interval)
{

    ts->nanoseconds  += interval->nanoseconds;
    ts->seconds      += interval->seconds;

    utils_time_interval_correct_of_uf(interval);
}

void utils_time_interval_add(time_interval_t *interval, time_interval_t *add)
{

    interval->nanoseconds  += add->nanoseconds;
    interval->seconds      += add->seconds;

    utils_time_interval_correct_of_uf(interval);
}

void utils_time_interval_sub(time_interval_t *interval, time_interval_t *sub)
{
    interval->nanoseconds  -= sub->nanoseconds;
    interval->seconds      -= sub->seconds;

    utils_time_interval_correct_of_uf(interval);
}

// use this function to correct eventually occured under and overflows in timer intervals
void utils_time_interval_correct_of_uf(time_interval_t *interval)
{
    if (interval->seconds != 0 && interval->nanoseconds < 0)
    {
        interval->seconds--;
        interval->nanoseconds += SECONDS_TO_NANOSECONDS;
    }
    else if(interval->nanoseconds <= -SECONDS_TO_NANOSECONDS)
    {
        interval->seconds--;
        interval->nanoseconds = -(interval->nanoseconds + SECONDS_TO_NANOSECONDS);
    }
    else if (interval->nanoseconds >= SECONDS_TO_NANOSECONDS)
    {
        interval->seconds++;
        interval->nanoseconds -= SECONDS_TO_NANOSECONDS;
    }
}

// use this function to correct eventually occured under and overflows in timer intervals
void utils_time_stamp_correct_of(time_stamp_t *ts)
{
    if (ts->nanoseconds >= SECONDS_TO_NANOSECONDS)
    {
        ts->seconds++;
        ts->nanoseconds -= SECONDS_TO_NANOSECONDS;
    }
}



#if __BYTE_ORDER__ == __ORDER_LITTLE_ENDIAN__ 

uint16_t utils_htons(uint16_t data)
{
    uint16_t network_data =  ((data & 0xFF) << 8) | ((data & 0xFF00) >> 8);
    //network_data |= (data & 0xFF00) >> 8;
    return network_data;
}

uint16_t utils_ntohs(uint16_t data)
{
    uint16_t network_data =  ((data & 0xFF) << 8) | ((data & 0xFF00) >> 8);
    //network_data |= (data & 0xFF00) >> 8;
    return network_data;
}

uint32_t utils_htonl(uint32_t data)
{
    uint32_t network_data = ((data & 0x000000FF) << 24) | ((data & 0x0000FF00) << 8) | ((data & 0x00FF0000) >> 8) | ((data & 0xFF000000) >> 24);
    /*
    network_data |= (data & 0x000000FF) << 24;
    network_data |= (data & 0x0000FF00) << 8;
    network_data |= (data & 0x00FF0000) >> 8;
    network_data |= (data & 0xFF000000) >> 24;
    */
    return network_data;
}

uint32_t utils_ntohl(uint32_t data)
{
    uint32_t network_data = ((data & 0x000000FF) << 24) | ((data & 0x0000FF00) << 8) | ((data & 0x00FF0000) >> 8) | ((data & 0xFF000000) >> 24);
    /*
    network_data |= (data & 0x000000FF) << 24;
    network_data |= (data & 0x0000FF00) << 8;
    network_data |= (data & 0x00FF0000) >> 8;
    network_data |= (data & 0xFF000000) >> 24;
    */
    return network_data;
}

#elif __BYTE_ORDER__ == __ORDER_BIG_ENDIAN__

uint16_t utils_htons(uint16_t data)
{
    return data;
}

uint16_t utils_ntohs(uint16_t data)
{
    return data;
}

uint32_t utils_htonl(uint32_t data)
{
    return data;
}

uint32_t utils_ntohl(uint32_t data)
{
    return data;
}
#else
    #error Macro __BYTE_ORDER is neither __LITTLE_ENDIAN nor __BIG_ENDIAN
#endif