#include <stdlib.h>
#include <stdio.h>

extern "C" {
    #include "../../inc/buffer_management.h"
    #include "../../library/buffer_management/tlsf.h"
    #include "../../inc/generic_types.h"
}

#include "CppUTest/TestHarness.h"


#define TEST_MEMORY_SPACE 2048
#define LOOPING_TESTS_ITEARTIONS 100
#define RANDOM_ITERATIONS_MAX 100000

static unsigned char *space;
ptr_t  tlsf_config_ptr;
uint16_t max_buffer;

uint16_t nlz_test(uint16_t x)
{ 
	uint16_t n = 16; 
	if ((x>>8 ) != 0) { n=n- 8; x = x>> 8; } 
	if ((x>>4 ) != 0) { n=n- 4; x = x>> 4; } 
	if ((x>>2 ) != 0) { n=n- 2; x = x>> 2; } 
	if ((x>>1 ) != 0) { return n-2; } 
	return n - x; 	
}

unsigned short get_tlsf_fli_sli(unsigned short size, tlsf_runtime_config_prt_t  cfg)
{
    uint16_t fli = 16 - nlz_test(size)-1;
	uint16_t shift_amount = (fli - cfg->sl_cls_cnt);
	uint16_t tmp = (size) >> shift_amount;
	uint16_t sli = tmp & ((1 << cfg->sl_cls_cnt)-1);
	fli -= cfg->fl_lowest_cls;

    return (fli << 8) | (sli & 0xFF); 
}

unsigned int check_free_lists(ptr_t config_ptr)
{
    tlsf_runtime_config_prt_t  cfg = (tlsf_runtime_config_prt_t) config_ptr;
    unsigned long memory_space = ((unsigned long) cfg->end_address) -((unsigned long) cfg->start_address);
    unsigned int non_empty_freelists=0;
    unsigned int sl_cls_cnt = 1 << cfg->sl_cls_cnt;

    for(unsigned int i=0; i<cfg->fl_cls_cnt; i++)
    {
        for( unsigned int j=0; j <  sl_cls_cnt; j++)
        {
            unsigned int free_list_id = i* sl_cls_cnt + j;
            unsigned int mask = (1 << j);
            if((cfg->sl_free[i] & mask) != 0 )
            {
                non_empty_freelists++;
                
                CHECK_TEXT(cfg->free_list_first[free_list_id] != NULL, "Mask shows free buffer, but not linked in free_list_first");
                CHECK_TEXT(cfg->free_list_last[free_list_id]  != NULL, "Mask shows free buffer, but not linked in free_list_last");
                block_header_free_tlsf_ptr_t buffer = cfg->free_list_first[free_list_id];
    
                // check if allocation marker is set correctly and size has a valid value
                CHECK_COMPARE(buffer->bh.alloc, ==, 0);
                CHECK_COMPARE(buffer->bh.size, >, 0);
                CHECK_COMPARE((unsigned long) (buffer->bh.size << 2), <, memory_space - sizeof(block_header_tlsf_t));
                CHECK_COMPARE(get_tlsf_fli_sli(buffer->bh.size << 2, cfg), !=, 0xFFFF);

                // check for ordering issues
                CHECK_TEXT(cfg->free_list_first[free_list_id] != buffer->next, "Buffer misordering in free_list_first");
                CHECK_TEXT(cfg->free_list_last[free_list_id]  != buffer->prev, "Buffer misordering in free_list_last");
                
                if(cfg->free_list_first[free_list_id] == cfg->free_list_last[free_list_id])
                {
                    // if the buffer is the only element in a free list, the next an previous pointer shall be NULL
                    POINTERS_EQUAL(NULL, buffer->prev);
                    POINTERS_EQUAL(NULL, buffer->next);
                }
            }
            else
            {
                POINTERS_EQUAL(NULL, cfg->free_list_first[free_list_id]);
                POINTERS_EQUAL(NULL, cfg->free_list_last[free_list_id]);
            }
        }
    }
    return non_empty_freelists;
}

unsigned int check_buffers(ptr_t config_ptr)
{
    tlsf_runtime_config_prt_t  cfg = (tlsf_runtime_config_prt_t) config_ptr;

    ptr_t buffer_address = (ptr_t) cfg->start_address;
	uint16_t lastBlockRead = 0;

	uint32_t buffer_cnt =0;
    uint8_t last_buffer_alloc = 1;
    uint16_t size_prev =0;
	while (lastBlockRead == 0)
	{

		block_header_free_tlsf_ptr_t buffer = (block_header_free_tlsf_ptr_t ) buffer_address;
        CHECK_COMPARE(buffer->bh.size, !=, 0);
        CHECK_COMPARE(buffer->bh.size << 2, >=, 1 << cfg->fl_lowest_cls);
        CHECK_TEXT((buffer->bh.alloc == 1) | (last_buffer_alloc == 1), "Two neighbouring buffers are free but not coalesced");
        CHECK_TEXT(buffer->bh.size_prev == size_prev, "Size of priveous buffer and size_prev do not match");

        CHECK_COMPARE(get_tlsf_fli_sli(buffer->bh.size << 2, cfg), !=, 0xFFFF);
        uint16_t fli = (get_tlsf_fli_sli(buffer->bh.size << 2, cfg) >> 8) & 0xFF;
        uint16_t sli =  get_tlsf_fli_sli(buffer->bh.size << 2, cfg)       & 0xFF;
        
		if(buffer->bh.alloc == 0)
		{
            unsigned int free_list_id = fli* (1 << cfg->sl_cls_cnt) + sli;

            CHECK_TEXT(cfg->free_list_first[free_list_id] != NULL, "Found free buffer, but free_list_first is empty");
            CHECK_TEXT(cfg->free_list_last[free_list_id]  != NULL, "Found free buffer, but free_list_last is empty");
            
            unsigned short fli_mask = (1 << fli);
            unsigned short sli_mask = (1 << sli);
            CHECK_COMPARE(cfg->fl_free & fli_mask, !=, 0);
            CHECK_COMPARE(cfg->sl_free[fli] & sli_mask, !=, 0);

            unsigned char buffer_found=0;
            
			block_header_free_tlsf_ptr_t  free_buffer = cfg->free_list_first[free_list_id];
            
            do
            {
                if((ptr_t)free_buffer == (ptr_t)buffer)
                {
                    buffer_found++;
                }
                free_buffer = (block_header_free_tlsf_ptr_t) free_buffer->next;
            }while(free_buffer != NULL);

            LONGS_EQUAL(1, buffer_found);
        }
		else
		{
            unsigned int buffer_found=0;
			for(unsigned int i=0; i<cfg->fl_cls_cnt; i++)
            {
                unsigned int sl_cls_cnt = (1 << cfg->sl_cls_cnt);
                for( unsigned int j=0; j < sl_cls_cnt ; j++)
                {
                    unsigned int free_list_id = i* sl_cls_cnt + j;
                    if(cfg->free_list_first[free_list_id] != NULL)
                    {
                        block_header_free_tlsf_ptr_t  free_buffer = cfg->free_list_first[free_list_id];
                        do
                        {
                            if((ptr_t)free_buffer == (ptr_t)buffer)
                            {
                                buffer_found++;
                            }
                            free_buffer = (block_header_free_tlsf_ptr_t) free_buffer->next;
                        }while(free_buffer != NULL);

                    }
                }
            }
            CHECK_TEXT(0 == buffer_found, "Buffer is allocated, but was found in free list");
		}
        
        last_buffer_alloc = buffer->bh.alloc;
        size_prev = buffer->bh.size;

		buffer_address += sizeof(block_header_tlsf_t) + (buffer->bh.size << 2); 	
        if(buffer->bh.last_block == 1)
		{
			CHECK_COMPARE(buffer_address, <=, cfg->end_address);//, "Last Block exceeds memory space");
            uint8_t* last_usable_address = (uint8_t*)cfg->end_address - 3;
            CHECK_COMPARE(buffer_address, ==, last_usable_address);//, "Last Block does not end with memory space");
            lastBlockRead = 0xFF;
		}
        else
        {
            CHECK_TEXT(buffer_address <= cfg->end_address, "Block exceeds memory space");
        }
        

        buffer_cnt++;
	}
    return buffer_cnt;
}

TEST_GROUP(TestTlsf) {
    
    void setup() {
        space = (unsigned char *) malloc(TEST_MEMORY_SPACE * sizeof(unsigned char));
        
        buffer_management_config_t init_config;
        init_config.start_address = space ;
        init_config.end_address = &space[TEST_MEMORY_SPACE-1];
        init_config.tlsf_cfg.sl_cls_cnt = 2;
        init_config.tlsf_cfg.fl_lowest_cls = 5;
        init_config.tlsf_cfg.fl_cls_cnt = 16 - init_config.tlsf_cfg.fl_lowest_cls;

        tlsf_config_ptr = tlsf_init(&init_config);
        tlsf_runtime_config_prt_t  cfg = (tlsf_runtime_config_prt_t) tlsf_config_ptr;

        max_buffer= (uint16_t) ((ptr_t)cfg->end_address - (ptr_t) cfg->start_address- sizeof(block_header_tlsf_t));
    }

    void teardown() {
        free(space);
    }
};

TEST(TestTlsf, Test_TlsfInit){
    tlsf_runtime_config_prt_t  cfg = (tlsf_runtime_config_prt_t) tlsf_config_ptr;

    if(cfg->fl_free == 0)
    {
        FAIL("No buffer found in first level, expected at least one");
    }
    
    unsigned int non_empty_free_lists = check_free_lists(tlsf_config_ptr);
    LONGS_EQUAL(1, non_empty_free_lists);
    
    unsigned int buffer_cnt= check_buffers(tlsf_config_ptr);
    LONGS_EQUAL(1, buffer_cnt);
    
}

TEST(TestTlsf, Test_TlsfGet){
    (void)tlsf_malloc(20, tlsf_config_ptr);
    unsigned int non_empty_free_lists = check_free_lists(tlsf_config_ptr);
    LONGS_EQUAL(1, non_empty_free_lists);

    unsigned int buffer_cnt= check_buffers(tlsf_config_ptr);
    LONGS_EQUAL(2, buffer_cnt);
}

TEST(TestTlsf, Test_TlsfGetFree){
    tlsf_runtime_config_prt_t  cfg = (tlsf_runtime_config_prt_t) tlsf_config_ptr;
    
    unsigned short  minimal_buffer = (1<<(cfg->fl_lowest_cls));
    ptr_t buffer = tlsf_malloc(minimal_buffer, tlsf_config_ptr);

    tlsf_free(buffer, tlsf_config_ptr);
    
    unsigned int non_empty_free_lists = check_free_lists(tlsf_config_ptr);
    LONGS_EQUAL(1, non_empty_free_lists);

    unsigned int buffer_cnt= check_buffers(tlsf_config_ptr);
    LONGS_EQUAL(1, buffer_cnt);
}

TEST(TestTlsf, Test_TlsfGetGetFree){
    tlsf_runtime_config_prt_t  cfg = (tlsf_runtime_config_prt_t) tlsf_config_ptr;
    
    unsigned short  minimal_buffer = (1<<(cfg->fl_lowest_cls));
    ptr_t buffer = tlsf_malloc(minimal_buffer, tlsf_config_ptr);
    (void) tlsf_malloc(minimal_buffer, tlsf_config_ptr);

    tlsf_free(buffer, tlsf_config_ptr);
    
    unsigned int non_empty_free_lists = check_free_lists(tlsf_config_ptr);
    LONGS_EQUAL(2, non_empty_free_lists);

    unsigned int buffer_cnt= check_buffers(tlsf_config_ptr);
    LONGS_EQUAL(3, buffer_cnt);
}

TEST(TestTlsf, Test_TlsfFull){

    tlsf_runtime_config_prt_t  cfg = (tlsf_runtime_config_prt_t) tlsf_config_ptr;
    
    unsigned short  minimal_buffer = (1<<(cfg->fl_lowest_cls));
    
    unsigned long acquired_buffer_cnt=0;

    uint8_t * next_buffer = NULL;

    // iterate until all buffers are allocated
    do{
        next_buffer = (uint8_t *) tlsf_malloc(minimal_buffer, tlsf_config_ptr);
        
        (void) check_free_lists(tlsf_config_ptr);
        (void) check_buffers(tlsf_config_ptr);

        if(next_buffer != NULL)
        {
            CHECK_TEXT(next_buffer + minimal_buffer < cfg->end_address, "Returned Buffer exceeds allocated memory space");
            CHECK_TEXT(next_buffer >= cfg->start_address, "Returned Buffer exceeds allocated memory space");
            
            acquired_buffer_cnt++;
        }
        
    }while (next_buffer != NULL );

    if(cfg->fl_free > 1)
    {
        FAIL("expected maximum one buffer");
    }

    unsigned int non_empty_free_lists = check_free_lists(tlsf_config_ptr);
    CHECK_COMPARE(non_empty_free_lists, <=,1);

    unsigned int found_buffer_cnt= check_buffers(tlsf_config_ptr);
    CHECK_COMPARE(found_buffer_cnt, >=, acquired_buffer_cnt);
    CHECK_COMPARE(found_buffer_cnt, <=, acquired_buffer_cnt+ 1);
}

TEST(TestTlsf, Test_TlsfFullFree_LowerSideMerge){

    tlsf_runtime_config_prt_t  cfg = (tlsf_runtime_config_prt_t) tlsf_config_ptr;
    
    unsigned short  minimal_buffer = (1<<(cfg->fl_lowest_cls));

    unsigned long memory_space = ((unsigned long) cfg->end_address) -((unsigned long) cfg->start_address);
    
    unsigned long max_buffer_cnt = memory_space / minimal_buffer+1;
    unsigned long buffer_cnt=0;

    ptr_t buffer_array[max_buffer_cnt];
    uint8_t * next_buffer = NULL;
    // iterate until all buffers are allocated
    do{
        next_buffer = (uint8_t *) tlsf_malloc(minimal_buffer, tlsf_config_ptr);
        
        (void) check_free_lists(tlsf_config_ptr);
        (void) check_buffers(tlsf_config_ptr);

        if(next_buffer != NULL)
        {
            CHECK_TEXT(next_buffer + minimal_buffer < cfg->end_address, "Returned Buffer exceeds memory space");
            CHECK_TEXT(next_buffer >= cfg->start_address, "Returned Buffer exceeds allocated memory space");
            
            buffer_array[buffer_cnt] = next_buffer;
            buffer_cnt++;
        }
    }while (next_buffer != NULL);
        
    for(unsigned int i=0; i< buffer_cnt ; i++)
    {
        tlsf_free(buffer_array[i], tlsf_config_ptr);
        (void) check_free_lists(tlsf_config_ptr);
        (void) check_buffers(tlsf_config_ptr);
    }
    unsigned int non_empty_free_lists = check_free_lists(tlsf_config_ptr);
    LONGS_EQUAL(1, non_empty_free_lists);

    unsigned int found_buffer_cnt= check_buffers(tlsf_config_ptr);
    CHECK_COMPARE(found_buffer_cnt, ==, 1);
}

TEST(TestTlsf, Test_TlsfFullFree_HigherSideMerge){

    tlsf_runtime_config_prt_t  cfg = (tlsf_runtime_config_prt_t) tlsf_config_ptr;
    
    unsigned short  minimal_buffer = (1<<(cfg->fl_lowest_cls));
    unsigned long memory_space = ((unsigned long) cfg->end_address) -((unsigned long) cfg->start_address);
    
    unsigned long max_buffer_cnt = memory_space / minimal_buffer+1;
    unsigned long buffer_cnt=0;

    ptr_t buffer_array[max_buffer_cnt];
    uint8_t * next_buffer = NULL;
    // iterate until all buffers are allocated
    do{
        next_buffer = (uint8_t *) tlsf_malloc(minimal_buffer, tlsf_config_ptr);
        
        (void) check_free_lists(tlsf_config_ptr);
        (void) check_buffers(tlsf_config_ptr);

        if(next_buffer != NULL)
        {
            CHECK_TEXT(next_buffer + minimal_buffer < cfg->end_address, "Returned Buffer exceeds memory space");
            CHECK_TEXT(next_buffer >= cfg->start_address, "Returned Buffer exceeds allocated memory space");
            
            buffer_array[buffer_cnt] = next_buffer;
            buffer_cnt++;
        }
    }while (next_buffer != NULL);
        
    for(unsigned int i=buffer_cnt-1; i< buffer_cnt; i--)
    {
        tlsf_free(buffer_array[i], tlsf_config_ptr);
        (void) check_free_lists(tlsf_config_ptr);
        (void) check_buffers(tlsf_config_ptr);
    }
    unsigned int non_empty_free_lists = check_free_lists(tlsf_config_ptr);
    LONGS_EQUAL(1, non_empty_free_lists);

    unsigned int found_buffer_cnt= check_buffers(tlsf_config_ptr);
    CHECK_COMPARE(found_buffer_cnt, ==, 1);
}

TEST(TestTlsf, Test_TlsfFullFree_DoubleSidedMerge){

    tlsf_runtime_config_prt_t  cfg = (tlsf_runtime_config_prt_t) tlsf_config_ptr;
    
    unsigned short  minimal_buffer = (1<<(cfg->fl_lowest_cls));
    unsigned long memory_space = ((unsigned long) cfg->end_address) -((unsigned long) cfg->start_address);
    
    unsigned long max_buffer_cnt = memory_space / minimal_buffer+1;
    unsigned long buffer_cnt=0;

    ptr_t buffer_array[max_buffer_cnt];
    uint8_t * next_buffer = NULL;
    // iterate until all buffers are allocated
    do{
        next_buffer = (uint8_t *)tlsf_malloc(minimal_buffer, tlsf_config_ptr);
        (void) check_free_lists(tlsf_config_ptr);
        (void) check_buffers(tlsf_config_ptr);

        if(next_buffer != NULL)
        {
            CHECK_TEXT(next_buffer + minimal_buffer < cfg->end_address, "Returned Buffer exceeds memory space");
            CHECK_TEXT(next_buffer >= cfg->start_address, "Returned Buffer exceeds allocated memory space");
            
            buffer_array[buffer_cnt] = next_buffer;
            buffer_cnt++;
        }
    }while (next_buffer != NULL);
        
    for(unsigned int i=0; i< buffer_cnt ; i+=2)
    {
        tlsf_free(buffer_array[i], tlsf_config_ptr);
        (void) check_free_lists(tlsf_config_ptr);
        (void) check_buffers(tlsf_config_ptr);
    }
    
    for(unsigned int i=1; i< buffer_cnt ; i+=2)
    {
        tlsf_free(buffer_array[i], tlsf_config_ptr);
        (void) check_free_lists(tlsf_config_ptr);
        (void) check_buffers(tlsf_config_ptr);
    }

    unsigned int non_empty_free_lists = check_free_lists(tlsf_config_ptr);
    LONGS_EQUAL(1, non_empty_free_lists);

    unsigned int found_buffer_cnt= check_buffers(tlsf_config_ptr);
    CHECK_COMPARE(found_buffer_cnt, ==, 1);
}

TEST(TestTlsf, Test_TlsfFullFree_RandomMerge){

    tlsf_runtime_config_prt_t  cfg = (tlsf_runtime_config_prt_t) tlsf_config_ptr;
    
    unsigned short  minimal_buffer = (1<<(cfg->fl_lowest_cls));
    unsigned long memory_space = ((unsigned long) cfg->end_address) -((unsigned long) cfg->start_address);
    
    unsigned long max_buffer_cnt = memory_space / minimal_buffer+1;
    unsigned long buffer_cnt=0;

    ptr_t buffer_array[max_buffer_cnt];
    uint8_t * next_buffer = NULL;
    // iterate until all buffers are allocated
    do{
        next_buffer = (uint8_t *) tlsf_malloc(minimal_buffer, tlsf_config_ptr);

        (void) check_free_lists(tlsf_config_ptr);
        (void) check_buffers(tlsf_config_ptr);

        if(next_buffer != NULL)
        {
            CHECK_TEXT(next_buffer + minimal_buffer < cfg->end_address, "Returned Buffer exceeds memory space");
            CHECK_TEXT(next_buffer >= cfg->start_address, "Returned Buffer exceeds allocated memory space");
            buffer_array[buffer_cnt] = next_buffer;
            buffer_cnt++;
        }
    }while (next_buffer != NULL);
    
    unsigned long buffers_left = buffer_cnt;
    while(buffers_left > 0)
    {
        unsigned long i = rand() % buffer_cnt;
        if(buffer_array[i] != NULL)
        {
            tlsf_free(buffer_array[i], tlsf_config_ptr);
            (void) check_free_lists(tlsf_config_ptr);
            (void) check_buffers(tlsf_config_ptr);
            buffer_array[i] = NULL;
            buffers_left--;
        }
    }

    unsigned int non_empty_free_lists = check_free_lists(tlsf_config_ptr);
    LONGS_EQUAL(1, non_empty_free_lists);

    unsigned int found_buffer_cnt= check_buffers(tlsf_config_ptr);
    CHECK_COMPARE(found_buffer_cnt, ==, 1);
}

TEST(TestTlsf, Test_TlsfFull_RandomSizes){

    tlsf_runtime_config_prt_t  cfg = (tlsf_runtime_config_prt_t) tlsf_config_ptr;
    
    unsigned short  minimal_buffer = (1<<(cfg->fl_lowest_cls));
    unsigned long memory_space = ((unsigned long) cfg->end_address) -((unsigned long) cfg->start_address);
    
    unsigned long max_buffer_cnt = memory_space / minimal_buffer+1;
    unsigned long acquired_buffer_cnt=0;

    unsigned long alloc_fails=0;
    unsigned long max_alloc_fails =100;

    uint8_t * next_buffer = NULL;

    // iterate until all buffers are allocated
    do{

        unsigned short buffer_size =  rand() % max_buffer;

        next_buffer = (uint8_t*) tlsf_malloc(buffer_size, tlsf_config_ptr);
        
        (void) check_free_lists(tlsf_config_ptr);
        (void) check_buffers(tlsf_config_ptr);

        if(next_buffer != NULL) 
        {
            CHECK_TEXT(next_buffer + buffer_size < cfg->end_address, "Returned Buffer exceeds memory space");
            CHECK_TEXT(next_buffer >= cfg->start_address, "Returned Buffer exceeds allocated memory space");
            
            acquired_buffer_cnt++;
            if(alloc_fails > 0)
            {
                alloc_fails--;
            }
        }
        else
        {
            // increment allocation fails if no buffer is available
            alloc_fails++;
        }
    }while (acquired_buffer_cnt < max_buffer_cnt && alloc_fails < max_alloc_fails);

    unsigned int non_empty_free_lists = check_free_lists(tlsf_config_ptr);
    CHECK_COMPARE(non_empty_free_lists, <=,1);

    unsigned int found_buffer_cnt= check_buffers(tlsf_config_ptr);
    CHECK_COMPARE(found_buffer_cnt, >=, acquired_buffer_cnt);
    CHECK_COMPARE(found_buffer_cnt, <=, acquired_buffer_cnt+ 1);
}

TEST(TestTlsf, Test_TlsfFullFree_RandomMergesAndSizes){

    tlsf_runtime_config_prt_t  cfg = (tlsf_runtime_config_prt_t) tlsf_config_ptr;
    
    unsigned short  minimal_buffer = (1<<(cfg->fl_lowest_cls));
    unsigned long memory_space = ((unsigned long) cfg->end_address) -((unsigned long) cfg->start_address);
    
    unsigned long max_buffer_cnt = memory_space / minimal_buffer+1;
    unsigned long acquired_buffer_cnt=0;

    unsigned long alloc_fails=0;
    unsigned long max_alloc_fails =100;

    uint8_t * buffer_array[max_buffer_cnt];
    uint8_t * next_buffer = NULL;

    // iterate until all buffers are allocated
    do{

        unsigned short buffer_size =  rand() % max_buffer;

        next_buffer = (uint8_t *)tlsf_malloc(buffer_size, tlsf_config_ptr);
        
        (void) check_free_lists(tlsf_config_ptr);
        (void) check_buffers(tlsf_config_ptr);

        if(next_buffer != NULL)
        {
            CHECK_TEXT(next_buffer + buffer_size < cfg->end_address, "Returned Buffer exceeds memory space");
            CHECK_TEXT(next_buffer >= cfg->start_address, "Returned Buffer exceeds allocated memory space");
            
            buffer_array[acquired_buffer_cnt] = next_buffer;
            acquired_buffer_cnt++;
            if(alloc_fails > 0)
            {
                alloc_fails--;
            }
        }
        else
        {
            // increment allocation fails if no buffer is available
            alloc_fails++;
        }
    }while (acquired_buffer_cnt < max_buffer_cnt && alloc_fails < max_alloc_fails);

    unsigned long buffers_left = acquired_buffer_cnt;
    while(buffers_left > 0)
    {
        unsigned long i = rand() % acquired_buffer_cnt;
        if(buffer_array[i] != NULL)
        {
            tlsf_free(buffer_array[i], tlsf_config_ptr);
            (void) check_free_lists(tlsf_config_ptr);
            (void) check_buffers(tlsf_config_ptr);
            buffer_array[i] = NULL;
            buffers_left--;
        }
    }

    unsigned int non_empty_free_lists = check_free_lists(tlsf_config_ptr);
    LONGS_EQUAL(1, non_empty_free_lists);

    unsigned int found_buffer_cnt= check_buffers(tlsf_config_ptr);
    CHECK_COMPARE(found_buffer_cnt, ==, 1);
}

TEST(TestTlsf, Test_TlsfFullFree_RandomMergesAndSizes_Looping){

    tlsf_runtime_config_prt_t  cfg = (tlsf_runtime_config_prt_t) tlsf_config_ptr;
    
    unsigned short  minimal_buffer = (1<<(cfg->fl_lowest_cls));
    unsigned long memory_space = ((unsigned long) cfg->end_address) -((unsigned long) cfg->start_address);
    
    unsigned long max_buffer_cnt = memory_space / minimal_buffer+1;
    unsigned long max_alloc_fails =100;
    for(unsigned int loopId=0; loopId < LOOPING_TESTS_ITEARTIONS; loopId++)
    {
        unsigned long acquired_buffer_cnt=0;
        unsigned long alloc_fails=0;

        ptr_t buffer_array[max_buffer_cnt];
        uint8_t * next_buffer = NULL;

        // iterate until all buffers are allocated  
        do{

            unsigned short buffer_size =  rand() % max_buffer;

            next_buffer = (uint8_t *)tlsf_malloc(buffer_size, tlsf_config_ptr);

            (void) check_free_lists(tlsf_config_ptr);
            (void) check_buffers(tlsf_config_ptr);

            if(next_buffer != NULL)
            {
                CHECK_TEXT(next_buffer + buffer_size < cfg->end_address, "Returned Buffer exceeds memory space");
                CHECK_TEXT(next_buffer >= cfg->start_address, "Returned Buffer exceeds allocated memory space");
                
                buffer_array[acquired_buffer_cnt] = next_buffer;
                acquired_buffer_cnt++;
                if(alloc_fails > 0)
                {
                    alloc_fails--;
                }
            }
            else
            {
                // increment allocation fails if no buffer is available
                alloc_fails++;
            }
        }while (acquired_buffer_cnt < max_buffer_cnt && alloc_fails < max_alloc_fails);

        unsigned long buffers_left = acquired_buffer_cnt;
        while(buffers_left > 0)
        {
            unsigned long i = rand() % acquired_buffer_cnt;
            if(buffer_array[i] != NULL)
            {
                tlsf_free(buffer_array[i], tlsf_config_ptr);
                buffer_array[i] = NULL;
                buffers_left--;
                (void) check_free_lists(tlsf_config_ptr);
                (void) check_buffers(tlsf_config_ptr);
            }
        }

        unsigned int non_empty_free_lists = check_free_lists(tlsf_config_ptr);
        LONGS_EQUAL(1, non_empty_free_lists);

        unsigned int found_buffer_cnt= check_buffers(tlsf_config_ptr);
        CHECK_COMPARE(found_buffer_cnt, ==, 1);
    }
}

TEST(TestTlsf, Test_Tlsf_RandomGetAndFree_MinimumSize){

    tlsf_runtime_config_prt_t  cfg = (tlsf_runtime_config_prt_t) tlsf_config_ptr;
    
    unsigned short  minimal_buffer = (1<<(cfg->fl_lowest_cls));
    unsigned long memory_space = ((unsigned long) cfg->end_address) -((unsigned long) cfg->start_address);
    
    unsigned long max_buffer_cnt = memory_space / minimal_buffer+1;
    unsigned long acquired_buffer_cnt=0;

    unsigned long alloc_fails=0;
    unsigned long max_alloc_fails =10;

    uint8_t * buffer_array[max_buffer_cnt];
    uint8_t * next_buffer = NULL;

    for(unsigned int i=0; i< max_buffer_cnt; i++)
    {
        buffer_array[i] = NULL;
    }
    unsigned long iterattions = 0;
    // iterate until all buffers are allocated
    do{
        unsigned short acquire = rand() %3;
        unsigned short buffer_id =  rand() % max_buffer_cnt;
        
        if(acquire == 1 && 
            buffer_array[buffer_id] == NULL &&
            acquired_buffer_cnt < max_buffer_cnt)
        {
            
            next_buffer = (uint8_t *)tlsf_malloc(minimal_buffer, tlsf_config_ptr);
            
            (void) check_free_lists(tlsf_config_ptr);
            (void) check_buffers(tlsf_config_ptr);

            if(next_buffer != NULL)
            {
                CHECK_TEXT(next_buffer + minimal_buffer < cfg->end_address, "Returned Buffer exceeds memory space");
                CHECK_TEXT(next_buffer >= cfg->start_address, "Returned Buffer exceeds allocated memory space");

                buffer_array[buffer_id] = next_buffer;
                acquired_buffer_cnt++;
                if(alloc_fails > 0)
                {
                    alloc_fails--;
                }
            }
            else
            {
                // increment allocation fails if no buffer is available
                alloc_fails++;
            }
        }
        else if(acquire == 0 &&
                buffer_array[buffer_id] != NULL)
        {            
            tlsf_free(buffer_array[buffer_id], tlsf_config_ptr);
                    
            buffer_array[buffer_id] = NULL;
            
            (void) check_free_lists(tlsf_config_ptr);
            (void) check_buffers(tlsf_config_ptr);
            
            acquired_buffer_cnt--;         
        } 
        
        iterattions++;
    }while (acquired_buffer_cnt < max_buffer_cnt && alloc_fails < max_alloc_fails && iterattions < RANDOM_ITERATIONS_MAX);

    unsigned long buffers_left = acquired_buffer_cnt;
    while(buffers_left > 0)
    {
        unsigned long i = rand() % max_buffer_cnt;
        if(buffer_array[i] != NULL)
        {
            tlsf_free(buffer_array[i], tlsf_config_ptr);
            (void) check_free_lists(tlsf_config_ptr);
            (void) check_buffers(tlsf_config_ptr);
            buffer_array[i] = NULL;
            buffers_left--;
        }
    }
    
    unsigned int non_empty_free_lists = check_free_lists(tlsf_config_ptr);
    LONGS_EQUAL(1, non_empty_free_lists);

    unsigned int found_buffer_cnt= check_buffers(tlsf_config_ptr);
    CHECK_COMPARE(found_buffer_cnt, ==, 1);
}

TEST(TestTlsf, Test_Tlsf_RandomGetAndFree_RandomSizes){

    tlsf_runtime_config_prt_t  cfg = (tlsf_runtime_config_prt_t) tlsf_config_ptr;
    
    unsigned short  minimal_buffer = (1<<(cfg->fl_lowest_cls));
    unsigned long memory_space = ((unsigned long) cfg->end_address) -((unsigned long) cfg->start_address);
    
    unsigned long max_buffer_cnt = memory_space / minimal_buffer+1;
    unsigned long acquired_buffer_cnt=0;

    unsigned long alloc_fails=0;
    unsigned long max_alloc_fails =1000;

    uint8_t * buffer_array[max_buffer_cnt];
    uint8_t * next_buffer = NULL;

    for(unsigned int i=0; i< max_buffer_cnt; i++)
    {
        buffer_array[i] = NULL;
    }
    unsigned long iterattions = 0;
    // iterate until all buffers are allocated
    do{
        unsigned short acquire = rand() %3;
        unsigned short buffer_id =  rand() % max_buffer_cnt;
        
        if(acquire == 1 && 
            buffer_array[buffer_id] == NULL &&
            acquired_buffer_cnt < max_buffer_cnt)
        {
            unsigned short buffer_size =  rand() % max_buffer;
            
            next_buffer = (uint8_t *)tlsf_malloc(buffer_size, tlsf_config_ptr);
            
            (void) check_free_lists(tlsf_config_ptr);
            (void) check_buffers(tlsf_config_ptr);

            if(next_buffer != NULL)
            {                
                block_header_free_tlsf_ptr_t  bl = (block_header_free_tlsf_ptr_t ) (next_buffer-sizeof(block_header_tlsf_t));
                CHECK_TEXT(bl->bh.alloc == 1, "Got unallocated buffer");
                CHECK_TEXT(next_buffer + buffer_size < cfg->end_address, "Returned Buffer exceeds memory space");
                CHECK_TEXT(next_buffer >= cfg->start_address, "Returned Buffer exceeds allocated memory space");
                
                buffer_array[buffer_id] = next_buffer;
                acquired_buffer_cnt++;
                if(alloc_fails > 0)
                {
                    alloc_fails--;
                }
            }
            else
            {
                // increment allocation fails if no buffer is available
                alloc_fails++;
            }
        }
        else if(acquire == 0 &&
                buffer_array[buffer_id] != NULL)
        {            
            next_buffer = buffer_array[buffer_id];
            next_buffer-= sizeof(block_header_tlsf_t);
            block_header_free_tlsf_ptr_t  bl = (block_header_free_tlsf_ptr_t ) next_buffer;
            CHECK_TEXT(bl->bh.alloc == 1, "Trying to free unallocated buffer");

            //tlsf_memory_dump(config);
            tlsf_free(buffer_array[buffer_id], tlsf_config_ptr);
                    
            buffer_array[buffer_id] = NULL;
            
            (void) check_free_lists(tlsf_config_ptr);
            (void) check_buffers(tlsf_config_ptr);
            
            acquired_buffer_cnt--;         
        } 
        
        iterattions++;
    }while (acquired_buffer_cnt < max_buffer_cnt && alloc_fails < max_alloc_fails && iterattions < RANDOM_ITERATIONS_MAX);

    unsigned long buffers_left = acquired_buffer_cnt;
    while(buffers_left > 0)
    {
        unsigned long i = rand() % max_buffer_cnt;
        if(buffer_array[i] != NULL)
        {
            tlsf_free(buffer_array[i], tlsf_config_ptr);
            (void) check_free_lists(tlsf_config_ptr);
            (void) check_buffers(tlsf_config_ptr);
            buffer_array[i] = NULL;
            buffers_left--;
        }
    }
    
    unsigned int non_empty_free_lists = check_free_lists(tlsf_config_ptr);
    LONGS_EQUAL(1, non_empty_free_lists);

    unsigned int found_buffer_cnt= check_buffers(tlsf_config_ptr);
    CHECK_COMPARE(found_buffer_cnt, ==, 1);
}

TEST(TestTlsf, Test_Tlsf_RandomGetAndFree_RandomSizes_FillBuffers){

    tlsf_runtime_config_prt_t  cfg = (tlsf_runtime_config_prt_t) tlsf_config_ptr;
    
    unsigned short  minimal_buffer = (1<<(cfg->fl_lowest_cls));
    unsigned long memory_space = ((unsigned long) cfg->end_address) -((unsigned long) cfg->start_address);
    
    unsigned long max_buffer_cnt = memory_space / minimal_buffer+1;
    unsigned long acquired_buffer_cnt=0;

    unsigned long alloc_fails=0;
    unsigned long max_alloc_fails =1000;

    uint8_t * buffer_array[max_buffer_cnt];
    uint8_t * next_buffer = NULL;

    for(unsigned int i=0; i< max_buffer_cnt; i++)
    {
        buffer_array[i] = NULL;
    }
    unsigned long iterattions = 0;
    // iterate until all buffers are allocated
    do{
        unsigned short acquire = rand() %3;
        unsigned short buffer_id =  rand() % max_buffer_cnt;
        
        if(acquire == 1 && 
            buffer_array[buffer_id] == NULL &&
            acquired_buffer_cnt < max_buffer_cnt)
        {
            unsigned short buffer_size =  rand() % max_buffer;
            
            next_buffer = (uint8_t *)tlsf_malloc(buffer_size, tlsf_config_ptr);
            (void) check_free_lists(tlsf_config_ptr);
            (void) check_buffers(tlsf_config_ptr);

            if(next_buffer != NULL)
            {
                CHECK_TEXT(next_buffer + buffer_size < cfg->end_address, "Returned Buffer exceeds memory space");
                CHECK_TEXT(next_buffer >= cfg->start_address, "Returned Buffer exceeds allocated memory space");
                
                buffer_array[buffer_id] = next_buffer;
                acquired_buffer_cnt++;
                if(alloc_fails > 0)
                {
                    alloc_fails--;
                }
                
                for(unsigned short i =0; i<buffer_size; i++)
                {
                    next_buffer[i] = rand() & 0xFF;
                }
                
            }
            else
            {
                // increment allocation fails if no buffer is available
                alloc_fails++;
            }
        }
        else if(acquire == 0 &&
                buffer_array[buffer_id] != NULL)
        {            
            tlsf_free(buffer_array[buffer_id], tlsf_config_ptr);

            buffer_array[buffer_id] = NULL;
            
            (void) check_free_lists(tlsf_config_ptr);
            (void) check_buffers(tlsf_config_ptr);
            
            acquired_buffer_cnt--;         
        } 
        
        iterattions++;
    }while (acquired_buffer_cnt < max_buffer_cnt && alloc_fails < max_alloc_fails && iterattions < RANDOM_ITERATIONS_MAX);

    unsigned long buffers_left = acquired_buffer_cnt;
    while(buffers_left > 0)
    {
        unsigned long i = rand() % max_buffer_cnt;
        if(buffer_array[i] != NULL)
        {
            tlsf_free(buffer_array[i], tlsf_config_ptr);
            (void) check_free_lists(tlsf_config_ptr);
            (void) check_buffers(tlsf_config_ptr);
            buffer_array[i] = NULL;
            buffers_left--;
        }
    }
    
    unsigned int non_empty_free_lists = check_free_lists(tlsf_config_ptr);
    LONGS_EQUAL(1, non_empty_free_lists);

    unsigned int found_buffer_cnt= check_buffers(tlsf_config_ptr);
    CHECK_COMPARE(found_buffer_cnt, ==, 1);
}

